# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Internal implementation of request Body validating middleware.

"""

import re

import jsonschema
from jsonschema import exceptions as jsonschema_exc
from oslo_utils import timeutils
from oslo_utils import uuidutils

from masakari.api.validation import parameter_types
from masakari import exception
from masakari.i18n import _


def _soft_validate_additional_properties(validator,
                                         additional_properties_value,
                                         instance,
                                         schema):
    """Validator function.

    If there are not any properties on the instance that are not specified in
    the schema, this will return without any effect. If there are any such
    extra properties, they will be handled as follows:

    - if the validator passed to the method is not of type "object", this
      method will return without any effect.
    - if the 'additional_properties_value' parameter is True, this method will
      return without any effect.
    - if the schema has an additionalProperties value of True, the extra
      properties on the instance will not be touched.
    - if the schema has an additionalProperties value of False and there
      aren't patternProperties specified, the extra properties will be stripped
      from the instance.
    - if the schema has an additionalProperties value of False and there
      are patternProperties specified, the extra properties will not be
      touched and raise validation error if pattern doesn't match.
    """
    if (not validator.is_type(instance, "object") or
            additional_properties_value):
        return

    properties = schema.get("properties", {})
    patterns = "|".join(schema.get("patternProperties", {}))
    extra_properties = set()
    for prop in instance:
        if prop not in properties:
            if patterns:
                if not re.search(patterns, prop):
                    extra_properties.add(prop)
            else:
                extra_properties.add(prop)

    if not extra_properties:
        return

    if patterns:
        error = "Additional properties are not allowed (%s %s unexpected)"
        if len(extra_properties) == 1:
            verb = "was"
        else:
            verb = "were"
        yield jsonschema_exc.ValidationError(
            error % (", ".join(repr(extra) for extra in extra_properties),
                     verb))
    else:
        for prop in extra_properties:
            del instance[prop]


@jsonschema.FormatChecker.cls_checks('date-time')
def _validate_datetime_format(instance):
    try:
        timeutils.parse_isotime(instance)
    except ValueError:
        return False
    else:
        return True


@jsonschema.FormatChecker.cls_checks('uuid')
def _validate_uuid_format(instance):
    return uuidutils.is_uuid_like(instance)


@jsonschema.FormatChecker.cls_checks('name', exception.InvalidName)
def _validate_name(instance):
    regex = parameter_types.valid_name_regex
    try:
        if re.search(regex.regex, instance):
            return True
    except TypeError:
        # The name must be string type. If instance isn't string type, the
        # TypeError will be raised at here.
        pass
    raise exception.InvalidName(reason=regex.reason)


class FormatChecker(jsonschema.FormatChecker):
    """A FormatChecker can output the message from cause exception

       We need understandable validation errors messages for users. When a
       custom checker has an exception, the FormatChecker will output a
       readable message provided by the checker.
    """

    def check(self, instance, format):
        """Check whether the instance conforms to the given format.

        :argument instance: the instance to check
        :type: any primitive type (str, number, bool)
        :argument str format: the format that instance should conform to
        :raises: :exc:`FormatError` if instance does not conform to format
        """

        if format not in self.checkers:
            return

        # For safety reasons custom checkers can be registered with
        # allowed exception types. Anything else will fall into the
        # default formatter.
        func, raises = self.checkers[format]
        result, cause = None, None

        try:
            result = func(instance)
        except raises as e:
            cause = e
        if not result:
            msg = "%r is not a %r" % (instance, format)
            raise jsonschema_exc.FormatError(msg, cause=cause)


class _SchemaValidator(object):
    """A validator class

    This class is changed from Draft4Validator to validate minimum/maximum
    value of a string number(e.g. '10'). This changes can be removed when
    we tighten up the API definition and the XML conversion.
    Also FormatCheckers are added for checking data formats which would be
    passed through masakari api commonly.

    """
    validator = None
    validator_org = jsonschema.Draft4Validator

    def __init__(self, schema, relax_additional_properties=False):
        validators = {
            'minimum': self._validate_minimum,
            'maximum': self._validate_maximum,
        }
        if relax_additional_properties:
            validators[
                'additionalProperties'] = _soft_validate_additional_properties

        validator_cls = jsonschema.validators.extend(self.validator_org,
                                                     validators)
        format_checker = FormatChecker()
        self.validator = validator_cls(schema, format_checker=format_checker)

    def validate(self, *args, **kwargs):
        try:
            self.validator.validate(*args, **kwargs)
        except jsonschema.ValidationError as ex:
            if isinstance(ex.cause, exception.InvalidName):
                detail = ex.cause.format_message()
            elif len(ex.path) > 0:
                detail = _("Invalid input for field/attribute %(path)s."
                           " Value: %(value)s. %(message)s") % {
                    'path': ex.path.pop(), 'value': ex.instance,
                    'message': ex.message
                }
            else:
                detail = ex.message
            raise exception.ValidationError(detail=detail)
        except TypeError as ex:
            # NOTE: If passing non string value to patternProperties parameter,
            #       TypeError happens. Here is for catching the TypeError.
            detail = str(ex)
            raise exception.ValidationError(detail=detail)

    def _number_from_str(self, instance):
        try:
            value = int(instance)
        except (ValueError, TypeError):
            try:
                value = float(instance)
            except (ValueError, TypeError):
                return None
        return value

    def _validate_minimum(self, validator, minimum, instance, schema):
        instance = self._number_from_str(instance)
        if instance is None:
            return
        return self.validator_org.VALIDATORS['minimum'](validator, minimum,
                                                        instance, schema)

    def _validate_maximum(self, validator, maximum, instance, schema):
        instance = self._number_from_str(instance)
        if instance is None:
            return
        return self.validator_org.VALIDATORS['maximum'](validator, maximum,
                                                        instance, schema)



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from oslo_config import cfg

path_opts = [
    cfg.StrOpt('pybasedir',
               default=os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                    '../../')),
               help='Directory where the masakari python module is installed'),
    cfg.StrOpt('bindir',
               default=os.path.join(sys.prefix, 'local', 'bin'),
               help='Directory where masakari binaries are installed'),
    cfg.StrOpt('state_path',
               default='$pybasedir',
               help="Top-level directory for maintaining masakari's state"),
]


def basedir_def(*args):
    """Return an uninterpolated path relative to $pybasedir."""
    return os.path.join('$pybasedir', *args)


def bindir_def(*args):
    """Return an uninterpolated path relative to $bindir."""
    return os.path.join('$bindir', *args)


def state_path_def(*args):
    """Return an uninterpolated path relative to $state_path."""
    return os.path.join('$state_path', *args)


def register_opts(conf):
    conf.register_opts(path_opts)


def list_opts():
    return {"DEFAULT": path_opts}



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys


class UUIDSentinels(object):
    def __init__(self):
        from oslo_utils import uuidutils
        self._uuid_module = uuidutils
        self._sentinels = {}

    def __getattr__(self, name):
        if name.startswith('_'):
            raise ValueError('Sentinels must not start with _')
        if name not in self._sentinels:
            self._sentinels[name] = self._uuid_module.generate_uuid()
        return self._sentinels[name]


sys.modules[__name__] = UUIDSentinels()



# Copyright (c) 2018 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from unittest import mock

from oslo_utils import timeutils
from oslo_versionedobjects import fixture

from masakari.notifications.objects import base as notification
from masakari.objects import base
from masakari.objects import fields
from masakari import test


class TestNotificationBase(test.NoDBTestCase):

    @base.MasakariObjectRegistry.register_if(False)
    class TestObject(base.MasakariObject):
        VERSION = '1.0'
        fields = {
            'field_1': fields.StringField(),
            'field_2': fields.IntegerField(),
            'not_important_field': fields.IntegerField(),
        }

    @base.MasakariObjectRegistry.register_if(False)
    class TestNotificationPayload(notification.NotificationPayloadBase):
        VERSION = '1.0'

        SCHEMA = {
            'field_1': ('source_field', 'field_1'),
            'field_2': ('source_field', 'field_2'),
        }

        fields = {
            'extra_field': fields.StringField(),  # filled by ctor
            'field_1': fields.StringField(),  # filled by the schema
            'field_2': fields.IntegerField(),   # filled by the schema
        }

        def populate_schema(self, source_field):
            super(TestNotificationBase.TestNotificationPayload,
                  self).populate_schema(source_field=source_field)

    @base.MasakariObjectRegistry.register_if(False)
    class TestNotificationPayloadEmptySchema(
        notification.NotificationPayloadBase):
        VERSION = '1.0'

        fields = {
            'extra_field': fields.StringField(),  # filled by ctor
        }

    @notification.notification_sample('test-update-1.json')
    @notification.notification_sample('test-update-2.json')
    @base.MasakariObjectRegistry.register_if(False)
    class TestNotification(notification.NotificationBase):
        VERSION = '1.0'
        fields = {
            'payload': fields.ObjectField('TestNotificationPayload')
        }

    @base.MasakariObjectRegistry.register_if(False)
    class TestNotificationEmptySchema(notification.NotificationBase):
        VERSION = '1.0'
        fields = {
            'payload': fields.ObjectField('TestNotificationPayloadEmptySchema')
        }

    fake_service = {
        'created_at': timeutils.utcnow().replace(microsecond=0),
        'updated_at': None,
        'deleted_at': None,
        'deleted': False,
        'id': 123,
        'host': 'fake-host',
        'binary': 'masakari-fake',
        'topic': 'fake-service-topic',
        'report_count': 1,
        'forced_down': False,
        'disabled': False,
        'disabled_reason': None,
        'last_seen_up': None,
        'version': 1}

    expected_payload = {
        'masakari_object.name': 'TestNotificationPayload',
        'masakari_object.data': {
            'extra_field': 'test string',
            'field_1': 'test1',
            'field_2': 42},
        'masakari_object.version': '1.0',
        'masakari_object.namespace': 'masakari'}

    def setUp(self):
        super(TestNotificationBase, self).setUp()
        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        self.publisher = notification.NotificationPublisher(
            context=mock_context, host='fake-host',
            binary='masakari-fake')

        self.my_obj = self.TestObject(field_1='test1',
                                      field_2=42,
                                      not_important_field=13)

        self.payload = self.TestNotificationPayload(
            extra_field='test string')
        self.payload.populate_schema(source_field=self.my_obj)

        self.notification = self.TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.EventNotificationAction.SEGMENT_CREATE,
                phase=fields.EventNotificationPhase.START),
            publisher=self.publisher,
            priority=fields.EventNotificationPriority.INFO,
            payload=self.payload)

    def _verify_notification(self, mock_notifier, mock_context,
                             expected_event_type,
                             expected_payload):
        mock_notifier.prepare.assert_called_once_with(
            publisher_id='masakari-fake:fake-host')
        mock_notify = mock_notifier.prepare.return_value.info
        self.assertTrue(mock_notify.called)
        self.assertEqual(mock_notify.call_args[0][0], mock_context)
        self.assertEqual(mock_notify.call_args[1]['event_type'],
                         expected_event_type)
        actual_payload = mock_notify.call_args[1]['payload']
        self.assertJsonEqual(expected_payload, actual_payload)

    @mock.patch('masakari.rpc.NOTIFIER')
    def test_emit_notification(self, mock_notifier):

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        self.notification.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='segment.create.start',
            expected_payload=self.expected_payload)

    @mock.patch('masakari.rpc.NOTIFIER')
    def test_emit_with_host_and_binary_as_publisher(self, mock_notifier):
        noti = self.TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.EventNotificationAction.SEGMENT_CREATE),
            publisher=notification.NotificationPublisher(
                host='fake-host', binary='masakari-fake'),
            priority=fields.EventNotificationPriority.INFO,
            payload=self.payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='segment.create',
            expected_payload=self.expected_payload)

    @mock.patch('masakari.rpc.NOTIFIER')
    def test_emit_event_type_without_phase(self, mock_notifier):
        noti = self.TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.EventNotificationAction.SEGMENT_CREATE),
            publisher=self.publisher,
            priority=fields.EventNotificationPriority.INFO,
            payload=self.payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='segment.create',
            expected_payload=self.expected_payload)

    @mock.patch('masakari.rpc.NOTIFIER')
    def test_not_possible_to_emit_if_not_populated(self, mock_notifier):
        non_populated_payload = self.TestNotificationPayload(
            extra_field='test string')
        noti = self.TestNotification(
            event_type=notification.EventType(
                object='test_object',
                action=fields.EventNotificationAction.SEGMENT_CREATE),
            publisher=self.publisher,
            priority=fields.EventNotificationPriority.INFO,
            payload=non_populated_payload)

        mock_context = mock.Mock()
        self.assertRaises(AssertionError, noti.emit, mock_context)
        self.assertFalse(mock_notifier.called)

    @mock.patch('masakari.rpc.NOTIFIER')
    def test_empty_schema(self, mock_notifier):
        non_populated_payload = self.TestNotificationPayloadEmptySchema(
            extra_field='test string')
        noti = self.TestNotificationEmptySchema(
            event_type=notification.EventType(
                object='test_object',
                action=fields.EventNotificationAction.SEGMENT_CREATE),
            publisher=self.publisher,
            priority=fields.EventNotificationPriority.INFO,
            payload=non_populated_payload)

        mock_context = mock.Mock()
        mock_context.to_dict.return_value = {}
        noti.emit(mock_context)

        self._verify_notification(
            mock_notifier,
            mock_context,
            expected_event_type='segment.create',
            expected_payload={
                'masakari_object.name': 'TestNotificationPayloadEmptySchema',
                'masakari_object.data': {'extra_field': 'test string'},
                'masakari_object.version': '1.0',
                'masakari_object.namespace': 'masakari'})

    def test_sample_decorator(self):
        self.assertEqual(2, len(self.TestNotification.samples))
        self.assertIn('test-update-1.json', self.TestNotification.samples)
        self.assertIn('test-update-2.json', self.TestNotification.samples)


class TestNotificationObjectVersions(test.NoDBTestCase):

    def test_notification_payload_version_depends_on_the_schema(self):
        @base.MasakariObjectRegistry.register_if(False)
        class TestNotificationPayload(notification.NotificationPayloadBase):
            VERSION = '1.0'

            SCHEMA = {
                'field_1': ('source_field', 'field_1'),
                'field_2': ('source_field', 'field_2'),
            }

            fields = {
                'extra_field': fields.StringField(),  # filled by ctor
                'field_1': fields.StringField(),  # filled by the schema
                'field_2': fields.IntegerField(),   # filled by the schema
            }

        checker = fixture.ObjectVersionChecker(
            {'TestNotificationPayload': (TestNotificationPayload,)})

        old_hash = checker.get_hashes(extra_data_func=get_extra_data)
        TestNotificationPayload.SCHEMA['field_3'] = ('source_field',
                                                     'field_3')
        new_hash = checker.get_hashes(extra_data_func=get_extra_data)

        self.assertNotEqual(old_hash, new_hash)


def get_extra_data(obj_class):
    extra_data = tuple()

    # Get the SCHEMA items to add to the fingerprint
    # if we are looking at a notification
    if issubclass(obj_class, notification.NotificationPayloadBase):
        schema_data = collections.OrderedDict(
            sorted(obj_class.SCHEMA.items()))

        extra_data += (schema_data,)

    return extra_data



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import fixture as config_fixture
from oslo_policy import opts as policy_opts

import masakari.conf
from masakari.conf import paths
from masakari import config

CONF = masakari.conf.CONF


class ConfFixture(config_fixture.Config):
    """Fixture to manage global conf settings."""
    def setUp(self):
        super(ConfFixture, self).setUp()
        self.conf.set_default('api_paste_config',
                              paths.state_path_def('etc/masakari/api-paste.ini'
                                                   ),
                              group='wsgi')

        config.parse_args([], default_config_files=[], configure_db=False,
                          init_rpc=False)
        self.conf.set_default('connection', "sqlite://", group='database')
        policy_opts.set_defaults(self.conf)



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the failover segment api."""

import copy
from unittest import mock

from oslo_utils import timeutils

from masakari.api import utils as api_utils
from masakari.compute import nova as nova_obj
from masakari.engine import rpcapi as engine_rpcapi
from masakari import exception
from masakari.ha import api as ha_api
from masakari import objects
from masakari.objects import base as obj_base
from masakari.objects import fields
from masakari.objects import host as host_obj
from masakari.objects import notification as notification_obj
from masakari.objects import segment as segment_obj
from masakari import test
from masakari.tests.unit.api.openstack import fakes
from masakari.tests.unit import fakes as fakes_data
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)


def _make_segment_obj(segment_dict):
    return segment_obj.FailoverSegment(**segment_dict)


def _make_host_obj(host_dict):
    return host_obj.Host(**host_dict)


def _make_notification_obj(notification_dict):
    return notification_obj.Notification(**notification_dict)


class FailoverSegmentAPITestCase(test.NoDBTestCase):
    """Test Case for failover segment api."""

    def setUp(self):
        super(FailoverSegmentAPITestCase, self).setUp()
        self.segment_api = ha_api.FailoverSegmentAPI()
        self.req = fakes.HTTPRequest.blank('/v1/segments',
                                           use_admin_context=True)
        self.context = self.req.environ['masakari.context']
        self.failover_segment = fakes_data.create_fake_failover_segment(
            name="segment1", id=1, description="something",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )
        self.exception_in_use = exception.FailoverSegmentInUse(
            uuid=self.failover_segment.uuid)

    def _fake_notification_workflow(self, exc=None):
        if exc:
            return exc

    def _assert_segment_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The failover segment objects were not equal")

    @mock.patch.object(segment_obj.FailoverSegmentList, 'get_all')
    def test_get_all(self, mock_get_all):
        fake_failover_segment = fakes_data.create_fake_failover_segment(
            name="segment2", id=2, description="something",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment_2
        )
        fake_failover_segment_list = [self.failover_segment,
                                     fake_failover_segment]
        mock_get_all.return_value = fake_failover_segment_list
        result = self.segment_api.get_all(self.context, filters=None,
                                          sort_keys=None,
                                          sort_dirs=None, limit=None,
                                          marker=None)
        for i in range(len(result)):
            self._assert_segment_data(fake_failover_segment_list[i],
                                      _make_segment_obj(result[i]))

    @mock.patch.object(segment_obj.FailoverSegmentList, 'get_all')
    def test_get_all_marker_not_found(self, mock_get_all):

        mock_get_all.side_effect = exception.MarkerNotFound(marker='123')

        self.assertRaises(exception.MarkerNotFound, self.segment_api.get_all,
                          self.context, filters=None, sort_keys=None,
                          sort_dirs=None, limit=None, marker='123')

    @mock.patch.object(segment_obj.FailoverSegmentList, 'get_all')
    def test_get_all_by_recovery_method(self, mock_get_all):
        filters = {'recovery_method': 'auto'}
        self.segment_api.get_all(self.context, filters=filters,
                                 sort_keys=None, sort_dirs=None,
                                 limit=None, marker=None)
        mock_get_all.assert_called_once_with(self.context, filters=filters,
                                             sort_keys=None, sort_dirs=None,
                                             limit=None, marker=None)

    @mock.patch.object(segment_obj.FailoverSegmentList, 'get_all')
    def test_get_all_invalid_sort_dir(self, mock_get_all):

        mock_get_all.side_effect = exception.InvalidInput(
            reason="Unknown sort direction, must be 'asc' or 'desc'")
        self.assertRaises(exception.InvalidInput, self.segment_api.get_all,
                          self.context, filters=None, sort_keys=None,
                          sort_dirs=['abcd'], limit=None, marker=None)

    @mock.patch.object(segment_obj, 'FailoverSegment')
    @mock.patch.object(segment_obj.FailoverSegment, 'create')
    def test_create(self, mock_segment_create, mock_segment_obj):
        segment_data = {"name": "segment1",
                        "service_type": "COMPUTE",
                        "recovery_method": "auto",
                        "description": "something"}
        mock_segment_obj.return_value = self.failover_segment
        mock_segment_obj.create = mock.Mock()
        result = self.segment_api.create_segment(self.context, segment_data)
        self._assert_segment_data(
            self.failover_segment, _make_segment_obj(result))

    @mock.patch.object(segment_obj.FailoverSegment, 'create')
    @mock.patch.object(api_utils, 'notify_about_segment_api')
    def test_segment_create_exception(self, mock_notify_about_segment_api,
                              mock_segment_create):
        segment_data = {"name": "segment1",
                        "service_type": "COMPUTE",
                        "recovery_method": "auto",
                        "description": "something"}
        e = exception.InvalidInput(reason="TEST")
        mock_segment_create.side_effect = e
        mock_notify_about_segment_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.segment_api.create_segment, self.context,
                          segment_data)
        action = fields.EventNotificationAction.SEGMENT_CREATE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_segment(self, mock_get_segment):

        mock_get_segment.return_value = self.failover_segment

        result = self.segment_api.get_segment(self.context,
                                              uuidsentinel.fake_segment)
        self._assert_segment_data(self.failover_segment, result)

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_segment_not_found(self, mock_get_segment):

        self.assertRaises(exception.FailoverSegmentNotFound,
                          self.segment_api.get_segment, self.context, '123')

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(segment_obj, 'FailoverSegment')
    @mock.patch.object(segment_obj.FailoverSegment, 'save')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_update(self, mock_get, mock_update, mock_segment_obj,
                    mock_is_under_recovery):
        segment_data = {"name": "segment1"}
        mock_get.return_value = self.failover_segment
        mock_segment_obj.return_value = self.failover_segment
        mock_segment_obj.update = mock.Mock()
        mock_is_under_recovery.return_value = False
        result = self.segment_api.update_segment(self.context,
                                                 uuidsentinel.fake_segment,
                                                 segment_data)
        self._assert_segment_data(self.failover_segment, result)

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(segment_obj.FailoverSegment, 'update')
    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_segment_update_exception(self, mock_get,
                                      mock_notify_about_segment_api,
                                      mock_segment_update,
                                      mock_is_under_recovery):
        mock_get.return_value = self.failover_segment
        segment_data = {"name": "segment1",
                        "service_type": "COMPUTE",
                        "recovery_method": "auto",
                        "description": "something"}
        e = exception.InvalidInput(reason="TEST")
        mock_segment_update.side_effect = e
        mock_is_under_recovery.return_value = False
        mock_notify_about_segment_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.segment_api.update_segment, self.context,
                          uuidsentinel.fake_segment, segment_data)
        action = fields.EventNotificationAction.SEGMENT_UPDATE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(exception, 'FailoverSegmentInUse')
    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(segment_obj, 'FailoverSegment')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_update_segment_under_recovery(self, mock_get, mock_segment_obj,
                    mock_is_under_recovery, mock_FailoverSegmentInUse):
        segment_data = {"name": "segment1"}
        mock_get.return_value = self.failover_segment
        mock_segment_obj.return_value = self.failover_segment
        mock_is_under_recovery.return_value = True
        mock_FailoverSegmentInUse.return_value = self.exception_in_use
        self.assertRaises(type(self.exception_in_use),
                          self.segment_api.update_segment,
                          self.context, uuidsentinel.fake_segment,
                          segment_data)

    @mock.patch.object(segment_obj.FailoverSegment, 'destroy')
    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_segment_delete(self, mock_get, mock_is_under_recovery,
                    mock_segment_destroy):
        mock_get.return_value = self.failover_segment
        mock_is_under_recovery.return_value = False
        self.segment_api.delete_segment(self.context,
                                        uuidsentinel.fake_segment)
        mock_segment_destroy.assert_called_once()

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(segment_obj.FailoverSegment, 'destroy')
    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_segment_delete_exception(self, mock_get,
                                      mock_notify_about_segment_api,
                                      mock_segment_destroy,
                                      mock_is_under_recovery):
        mock_get.return_value = self.failover_segment
        mock_is_under_recovery.return_value = False
        e = exception.InvalidInput(reason="TEST")
        mock_segment_destroy.side_effect = e
        mock_notify_about_segment_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.segment_api.delete_segment, self.context,
                          uuidsentinel.fake_segment)

        action = fields.EventNotificationAction.SEGMENT_DELETE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)


class HostAPITestCase(test.NoDBTestCase):
    """Test Case for host api."""

    def setUp(self):
        super(HostAPITestCase, self).setUp()
        self.host_api = ha_api.HostAPI()
        self.req = fakes.HTTPRequest.blank(
            '/v1/segments/%s/hosts' % uuidsentinel.fake_segment,
            use_admin_context=True)
        self.context = self.req.environ['masakari.context']
        self.failover_segment = fakes_data.create_fake_failover_segment(
            name="segment1", id=1, description="something",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )
        self.host = fakes_data.create_fake_host(
            name="host_1", id=1, reserved=False, on_maintenance=False,
            type="fake", control_attributes="fake-control_attributes",
            uuid=uuidsentinel.fake_host_1
        )
        self.exception_in_use = exception.HostInUse(
            uuid=self.host.uuid)

    def _assert_host_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The host objects were not equal")

    @mock.patch.object(host_obj.HostList, 'get_all')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_all(self, mock_get, mock_get_all):
        mock_get.return_value = self.failover_segment
        fake_host = fakes_data.create_fake_host(
            name="host_2", id=2, reserved=False, on_maintenance=False,
            type="fake", control_attributes="fake-control_attributes",
            uuid=uuidsentinel.fake_host_2
        )
        fake_host_list = [self.host, fake_host]
        mock_get_all.return_value = fake_host_list

        result = self.host_api.get_all(self.context,
                                       filters=None, sort_keys=['created_at'],
                                       sort_dirs=['desc'], limit=None,
                                       marker=None)
        for i in range(len(result)):
            self._assert_host_data(
                fake_host_list[i], _make_host_obj(result[i]))

    @mock.patch.object(host_obj.HostList, 'get_all')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_all_marker_not_found(self, mock_get, mock_get_all):
        mock_get.return_value = self.failover_segment
        mock_get_all.side_effect = exception.MarkerNotFound(marker="1234")

        self.assertRaises(exception.MarkerNotFound, self.host_api.get_all,
                          self.context, filters=None, sort_keys=['created_at'],
                          sort_dirs=['desc'], limit=None,
                          marker="1234")

    @mock.patch.object(host_obj.HostList, 'get_all')
    def test_get_all_by_type(self, mock_get):
        filters = {'type': 'SSH',
                   'failover_segment_id': uuidsentinel.fake_segment}
        self.host_api.get_all(self.context, filters, sort_keys='created_at',
                              sort_dirs='desc', limit=None, marker=None)
        mock_get.assert_called_once_with(self.context, filters=filters,
                                         sort_keys='created_at',
                                         sort_dirs='desc',
                                         limit=None, marker=None)

    @mock.patch.object(host_obj.HostList, 'get_all')
    def test_get_all_invalid_sort_dir(self, mock_get):

        mock_get.side_effect = exception.InvalidInput(
            reason="Unknown sort direction, must be 'asc' or 'desc'")

        self.assertRaises(exception.InvalidInput, self.host_api.get_all,
                          self.context, filters=None, sort_keys=None,
                          sort_dirs=['abcd'], limit=None,
                          marker=None)

    @mock.patch.object(host_obj, 'Host')
    @mock.patch.object(host_obj.Host, 'create')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_create(self, mock_get, mock_find_compute_service,
                mock_host_create, mock_host_obj):
        mock_get.return_value = self.failover_segment

        host_data = {
            "name": 'host-1', "type": "fake-type",
            "reserved": False,
            "on_maintenance": False,
            "control_attributes": "fake-control_attributes"
        }
        mock_host_obj.return_value = self.host
        mock_host_obj.create = mock.Mock()
        result = self.host_api.create_host(self.context,
                                           uuidsentinel.fake_segment1,
                                           host_data)
        self._assert_host_data(self.host, _make_host_obj(result))

    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_create_non_existing_host(self, mock_segment_get,
                                    mock_find_compute_service):
        mock_segment_get.return_value = self.failover_segment
        mock_find_compute_service.side_effect = exception\
            .ComputeNotFoundByName(compute_name='host-2')

        host_data = {
            "name": 'host-2',
            "type": "fake-type",
            "reserved": False,
            "on_maintenance": False,
            "control_attributes": "fake-control_attributes"
        }

        self.assertRaises(exception.ComputeNotFoundByName,
                          self.host_api.create_host,
                          self.context, uuidsentinel.fake_segment1, host_data)

    @mock.patch.object(host_obj, 'Host')
    @mock.patch.object(host_obj.Host, 'create')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_host_create_exception(self, mock_get,
                                   mock_notify_about_host_api,
                                   mock_find_compute_service, mock_host_obj,
                                   mock_host_create):
        mock_get.return_value = self.failover_segment
        host_data = {
            "name": "host-1", "type": "fake-type",
            "reserved": False,
            "on_maintenance": False,
            "control_attributes": "fake-control_attributes"
        }

        e = exception.InvalidInput(reason="TEST")
        mock_host_obj.side_effect = e
        mock_notify_about_host_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.host_api.create_host, self.context,
                          uuidsentinel.fake_segment1, host_data)
        action = fields.EventNotificationAction.HOST_CREATE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)
        mock_find_compute_service.assert_called_once()

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('masakari.db.host_create')
    @mock.patch.object(host_obj.Host, '_from_db_object')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_create_convert_boolean_attributes(self, mock_get_segment,
                                               mock_find_compute_service,
                                               mock__from_db_object,
                                               mock_host_create,
                                               mock_generate_uuid,
                                               mock_notify_about_host_api):
        host_data = {
            "name": "host-1", "type": "fake-type",
            "reserved": 'On',
            "on_maintenance": '0',
            "control_attributes": "fake-control_attributes"
        }

        expected_data = {
            'reserved': True, 'name': 'host-1',
            'control_attributes': 'fake-control_attributes',
            'on_maintenance': False,
            'uuid': uuidsentinel.fake_uuid,
            'failover_segment_id': self.failover_segment.uuid,
            'type': 'fake-type'
        }
        mock_host_create.create = mock.Mock()
        mock_get_segment.return_value = self.failover_segment
        mock_generate_uuid.return_value = uuidsentinel.fake_uuid
        result = self.host_api.create_host(self.context,
                                           uuidsentinel.fake_segment1,
                                           host_data)
        mock_host_create.assert_called_with(self.context, expected_data)
        action = fields.EventNotificationAction.HOST_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, result, action=action,
                      phase=phase_start),
            mock.call(self.context, result, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_host(self, mock_get, mock_get_host):
        mock_get_host.return_value = self.host
        mock_get.return_value = self.failover_segment
        result = self.host_api.get_host(self.context,
                                        uuidsentinel.fake_segment,
                                        uuidsentinel.fake_host_1)
        self._assert_host_data(self.host, result)

    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    def test_get_host_not_found(self, mock_get, mock_get_host):
        self.assertRaises(exception.HostNotFound,
                          self.host_api.get_host, self.context,
                          uuidsentinel.fake_segment,
                          "123")

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(host_obj, 'Host')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(host_obj.Host, 'save')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_update(
            self, mock_get, mock_update, mock_find_compute_service,
            mock_host_obj, mock_is_under_recovery):
        host_data = {"name": "host_1"}
        mock_get.return_value = self.host
        mock_host_obj.return_value = self.host
        mock_host_obj.update = mock.Mock()
        mock_is_under_recovery.return_value = False
        result = self.host_api.update_host(self.context,
                                           uuidsentinel.fake_segment,
                                           uuidsentinel.fake_host_1,
                                           host_data)
        self._assert_host_data(self.host, result)

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_update_with_non_existing_host(
            self, mock_get, mock_find_compute_service,
            mock_is_under_recovery):
        host_data = {"name": "host-2"}
        mock_get.return_value = self.host
        mock_find_compute_service.side_effect = (
            exception.ComputeNotFoundByName(compute_name='host-2'))
        mock_is_under_recovery.return_value = False
        self.assertRaises(exception.ComputeNotFoundByName,
                          self.host_api.update_host, self.context,
                          uuidsentinel.fake_segment,
                          uuidsentinel.fake_host_1,
                          host_data)

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(nova_obj.API, 'find_compute_service')
    @mock.patch.object(host_obj.Host, 'save')
    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_host_update_exception(
            self, mock_get, mock_notify_about_host_api, mock_host_obj,
            mock_find_compute_service, mock_is_under_recovery):
        host_data = {"name": "host_1"}
        mock_get.return_value = self.host
        e = exception.InvalidInput(reason="TEST")
        mock_host_obj.side_effect = e
        mock_is_under_recovery.return_value = False
        mock_notify_about_host_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.host_api.update_host, self.context,
                          uuidsentinel.fake_segment,
                          uuidsentinel.fake_host_1,
                          host_data)
        action = fields.EventNotificationAction.HOST_UPDATE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(exception, 'HostInUse')
    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(host_obj, 'Host')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_update_host_under_recovery(
            self, mock_get, mock_host_obj, mock_is_under_recovery,
            mock_HostInUse):
        host_data = {"name": "host_1"}
        mock_get.return_value = self.host
        mock_host_obj.return_value = self.host
        mock_is_under_recovery.return_value = True
        mock_HostInUse.return_value = self.exception_in_use
        self.assertRaises(type(self.exception_in_use),
                          self.host_api.update_host,
                          self.context, uuidsentinel.fake_segment,
                          uuidsentinel.fake_host_1, host_data)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch('masakari.objects.base.MasakariObject'
                '.masakari_obj_get_changes')
    @mock.patch.object(host_obj.Host, '_from_db_object')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    @mock.patch('masakari.db.host_update')
    def test_update_convert_boolean_attributes(
            self, mock_host_update, mock_host_object, mock__from_db_object,
            mock_masakari_obj_get_changes, mock_is_under_recovery,
            mock_notify_about_host_api):
        host_data = {
            "reserved": 'Off',
            "on_maintenance": 'True',
        }

        expected_data = {
            'name': 'host_1', 'uuid': uuidsentinel.fake_host_1,
            'on_maintenance': True,
            'failover_segment': self.failover_segment,
            'reserved': False, 'type': 'fake',
            'control_attributes': 'fake-control_attributes'
        }

        mock_masakari_obj_get_changes.return_value = host_data
        mock_host_object.return_value = self.host
        self.host._context = self.context
        mock_is_under_recovery.return_value = False
        result = self.host_api.update_host(self.context,
                                           uuidsentinel.fake_segment,
                                           uuidsentinel.fake_host_1,
                                           host_data)
        mock_host_update.assert_called_with(self.context,
                                            uuidsentinel.fake_host_1,
                                            host_data)
        mock_host_update.return_value = expected_data
        action = fields.EventNotificationAction.HOST_UPDATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, result, action=action,
                      phase=phase_start),
            mock.call(self.context, result, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(host_obj.Host, 'destroy')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_delete_host(
            self, mock_get, mock_segment_destroy, mock_is_under_recovery):
        mock_get.return_value = self.host
        mock_is_under_recovery.return_value = False

        self.host_api.delete_host(self.context,
                                  uuidsentinel.fake_segment,
                                  uuidsentinel.fake_host_1)
        mock_segment_destroy.assert_called_once()

    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(host_obj.Host, 'destroy')
    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_host_delete_exception(
            self, mock_get, mock_notify_about_host_api, mock_host_destroy,
            mock_is_under_recovery):
        mock_get.return_value = self.host
        mock_is_under_recovery.return_value = False
        e = exception.InvalidInput(reason="TEST")
        mock_host_destroy.side_effect = e
        mock_notify_about_host_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.host_api.delete_host, self.context,
                          uuidsentinel.fake_segment,
                          uuidsentinel.fake_host_1)

        action = fields.EventNotificationAction.HOST_DELETE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(exception, 'HostInUse')
    @mock.patch.object(segment_obj.FailoverSegment,
                       'is_under_recovery')
    @mock.patch.object(host_obj, 'Host')
    @mock.patch.object(host_obj.Host, 'get_by_uuid')
    def test_delete_host_under_recovery(
            self, mock_get, mock_host_obj, mock_is_under_recovery,
            mock_HostInUse):
        mock_get.return_value = self.host
        mock_is_under_recovery.return_value = True
        mock_HostInUse.return_value = self.exception_in_use
        self.assertRaises(type(self.exception_in_use),
                          self.host_api.delete_host,
                          self.context, uuidsentinel.fake_segment,
                          uuidsentinel.fake_host_1)


class NotificationAPITestCase(test.NoDBTestCase):
    """Test Case for notification api."""

    @mock.patch.object(engine_rpcapi, 'EngineAPI')
    def setUp(self, mock_rpc):
        super(NotificationAPITestCase, self).setUp()
        self.notification_api = ha_api.NotificationAPI()
        self.req = fakes.HTTPRequest.blank('/v1/notifications',
                                           use_admin_context=True)
        self.context = self.req.environ['masakari.context']
        self.failover_segment = fakes_data.create_fake_failover_segment(
            name="segment1", id=1, description="something",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )
        self.host = fakes_data.create_fake_host(
            name="host_1", id=1, reserved=False, on_maintenance=False,
            type="fake", control_attributes="fake-control_attributes",
            uuid=uuidsentinel.fake_host_1
        )
        self.notification = fakes_data.create_fake_notification(
            type="VM", id=1, payload={
                'event': 'STOPPED', 'host_status': 'NORMAL',
                'cluster_status': 'ONLINE'
            },
            source_host_uuid=uuidsentinel.fake_host, generated_time=NOW,
            status="running",
            notification_uuid=uuidsentinel.fake_notification
        )
        self.exception_duplicate = exception.DuplicateNotification(
            host='host_1', type='COMPUTE_HOST')

    def _assert_notification_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The notification objects were not equal")

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    @mock.patch.object(notification_obj, 'Notification')
    @mock.patch.object(notification_obj.Notification, 'create')
    @mock.patch.object(host_obj.Host, 'get_by_name')
    def test_create(self, mock_host_obj, mock_create, mock_notification_obj,
                    mock_get_all):
        fake_notification = fakes_data.create_fake_notification(
            type="COMPUTE_HOST", id=2, payload={
                'event': 'STARTED', 'host_status': 'NORMAL',
                'cluster_status': 'ONLINE'
            },
            source_host_uuid=uuidsentinel.fake_host, generated_time=NOW,
            status="running",
            notification_uuid=uuidsentinel.fake_notification_2
        )
        fake_notification_list = [self.notification, fake_notification]
        mock_get_all.return_value = fake_notification_list
        notification_data = {"hostname": "fake_host",
                             "payload": {"event": "STARTED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "OFFLINE"},
                             "type": "VM",
                             "generated_time": "2016-10-13T09:11:21.656788"}
        mock_host_obj.return_value = self.host
        mock_notification_obj.return_value = self.notification

        result = (self.notification_api.
                  create_notification(self.context, notification_data))

        self._assert_notification_data(
            self.notification, _make_notification_obj(result))

    @mock.patch.object(api_utils, 'notify_about_notification_api')
    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    @mock.patch.object(notification_obj.Notification, 'create')
    @mock.patch.object(host_obj.Host, 'get_by_name')
    def test_create_notification_exception(self, mock_host_obj,
                                           mock_notification_obj, mock_get_all,
                                           mock_notify_about_notification_api):
        fake_notification = fakes_data.create_fake_notification(
            type="COMPUTE_HOST", id=2, payload={
                'event': 'STARTED', 'host_status': 'NORMAL',
                'cluster_status': 'ONLINE'
            },
            source_host_uuid=uuidsentinel.fake_host, generated_time=NOW,
            status="running",
            notification_uuid=uuidsentinel.fake_notification_2
        )
        fake_notification_list = [self.notification, fake_notification]
        mock_get_all.return_value = fake_notification_list
        notification_data = {"hostname": "fake_host",
                             "payload": {"event": "STARTED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "OFFLINE"},
                             "type": "VM",
                             "generated_time": "2016-10-13T09:11:21.656788"}
        mock_host_obj.return_value = self.host
        e = exception.InvalidInput(reason="TEST")
        mock_notification_obj.side_effect = e
        mock_notify_about_notification_api.return_value = mock.Mock()

        self.assertRaises(exception.InvalidInput,
                          self.notification_api.create_notification,
                          self.context, notification_data)

        action = fields.EventNotificationAction.NOTIFICATION_CREATE
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, mock.ANY, action=action,
                      phase=phase_error,
                      exception=e,
                      tb=mock.ANY)]
        mock_notify_about_notification_api.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, 'get_by_name')
    def test_create_host_on_maintenance(self, mock_host):
        self.host.on_maintenance = True
        mock_host.return_value = self.host
        notification_data = {"hostname": "host_1",
                             "payload": {"event": "STOPPED",
                                         "host_status": "NORMAL",
                                         "cluster_status": "ONLINE"},
                             "type": "COMPUTE_HOST",
                             "generated_time": str(NOW)}

        self.assertRaises(exception.HostOnMaintenanceError,
                          self.notification_api.create_notification,
                          self.context, notification_data)

    @mock.patch.object(exception, 'DuplicateNotification')
    @mock.patch.object(objects, 'Notification')
    @mock.patch.object(host_obj.Host, 'get_by_name')
    def test_create_duplicate_notification(self, mock_host,
        mock_notification_obj, mock_DuplicateNotification):
        mock_host.return_value = self.host
        self.notification_api._is_duplicate_notification = mock.Mock(
            return_value=True)
        notification_data = {"hostname": "host_1",
                             "payload": {'event': 'STOPPED',
                                         'host_status': 'NORMAL',
                                         'cluster_status': 'ONLINE'},
                             "type": "COMPUTE_HOST",
                             "generated_time": str(NOW)}
        self.notification.type = notification_data.get('type')
        self.notification.generated_time = notification_data.get(
            'generated_time')
        self.notification.source_host_uuid = self.host.uuid
        self.notification.payload = notification_data.get('payload')
        self.notification.status = fields.NotificationStatus.NEW
        mock_notification_obj.return_value = self.notification
        mock_DuplicateNotification.return_value = self.exception_duplicate

        self.assertRaises(type(self.exception_duplicate),
                          self.notification_api.create_notification,
                          self.context, notification_data)

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_create_is_duplicate_true(self, mock_get_all):
        mock_get_all.return_value = [self.notification, ]

        self.assertTrue(self.notification_api._is_duplicate_notification(
            self.context, self.notification))

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_is_duplicate_true_for_any_notification_status(
            self, mock_get_all):
        FAKE_NOTIFICATION_FINISHED = copy.deepcopy(self.notification)
        FAKE_NOTIFICATION_FINISHED.status = fields.NotificationStatus.FINISHED
        mock_get_all.return_value = [FAKE_NOTIFICATION_FINISHED]
        FAKE_NOTIFICATION_NEW = copy.deepcopy(self.notification)
        FAKE_NOTIFICATION_NEW.status = fields.NotificationStatus.NEW
        self.assertTrue(self.notification_api._is_duplicate_notification(
            self.context, FAKE_NOTIFICATION_NEW))

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_create_is_duplicate_false(self, mock_get_all):
        mock_get_all.return_value = [self.notification, ]
        FAKE_NOTIFICATION = copy.deepcopy(self.notification)
        FAKE_NOTIFICATION.payload = {'event': 'STOPPED',
                                     'host_status': 'UNKNOWN',
                                     'cluster_status': 'OFFLINE'}
        self.assertFalse(self.notification_api._is_duplicate_notification(
            self.context, FAKE_NOTIFICATION))

    @mock.patch.object(notification_obj.Notification, 'get_by_uuid')
    def test_get_notification(self, mock_get_notification):

        mock_get_notification.return_value = self.notification

        result = (self.notification_api.
                  get_notification(self.context,
                                   uuidsentinel.fake_notification))
        self._assert_notification_data(self.notification, result)

    @mock.patch.object(notification_obj.Notification, 'get_by_uuid')
    def test_get_notification_not_found(self, mock_get_notification):

        self.assertRaises(exception.NotificationNotFound,
                          self.notification_api.get_notification,
                          self.context, '123')

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_get_all(self, mock_get_all):
        fake_notification = fakes_data.create_fake_notification(
            type="VM", id=2, payload={
                'event': 'STOPPED', 'host_status': 'NORMAL',
                'cluster_status': 'ONLINE'
            },
            source_host_uuid=uuidsentinel.fake_host, generated_time=NOW,
            status="running",
            notification_uuid=uuidsentinel.fake_notification_2
        )
        fake_notification_list = [self.notification, fake_notification]
        mock_get_all.return_value = fake_notification_list

        result = self.notification_api.get_all(self.context, self.req)
        for i in range(len(result)):
            self._assert_notification_data(
                fake_notification_list[i], result[i])

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_get_all_marker_not_found(self, mock_get_all):

        mock_get_all.side_effect = exception.MarkerNotFound(marker="100")
        self.req = fakes.HTTPRequest.blank('/v1/notifications?marker=100',
                                           use_admin_context=True)
        self.assertRaises(exception.MarkerNotFound,
                          self.notification_api.get_all,
                          self.context, self.req)

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_get_all_by_status(self, mock_get_all):
        self.req = fakes.HTTPRequest.blank('/v1/notifications?status=new',
                                           use_admin_context=True)
        self.notification_api.get_all(self.context, filters={'status': 'new'},
                                      sort_keys='generated_time',
                                      sort_dirs='asc', limit=1000, marker=None)
        mock_get_all.assert_called_once_with(self.context, {'status': 'new'},
                                             'generated_time', 'asc',
                                             1000, None)

    @mock.patch.object(notification_obj.NotificationList, 'get_all')
    def test_get_all_invalid_sort_dir(self, mock_get_all):

        mock_get_all.side_effect = exception.InvalidInput(
            reason="Unknown sort direction, must be 'asc' or 'desc'")
        self.req = fakes.HTTPRequest.blank('/v1/notifications?sort_dir=abcd',
                                           use_admin_context=True)
        self.assertRaises(exception.InvalidInput,
                          self.notification_api.get_all,
                          self.context, self.req)



#    Copyright 2016 NTT DATA
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import inspect
import os
import pprint
from unittest import mock

from oslo_versionedobjects import exception as ovo_exc
from oslo_versionedobjects import fixture

from masakari import objects
from masakari.objects import base
from masakari.objects import fields
from masakari.objects import segment
from masakari import test
from masakari.tests.unit.objects import fake_args


class MyOwnedObject(base.MasakariPersistentObject, base.MasakariObject):
    VERSION = '1.0'
    fields = {'baz': fields.IntegerField()}


class MyObj(base.MasakariPersistentObject, base.MasakariObject,
            base.MasakariObjectDictCompat):
    VERSION = '1.6'
    fields = {'foo': fields.IntegerField(default=1),
              'bar': fields.StringField(),
              'missing': fields.StringField(),
              'readonly': fields.IntegerField(read_only=True),
              'rel_object': fields.ObjectField('MyOwnedObject', nullable=True),
              'rel_objects': fields.ListOfObjectsField('MyOwnedObject',
                                                       nullable=True),
              'mutable_default': fields.ListOfStringsField(default=[]),
              }

    @staticmethod
    def _from_db_object(context, obj, db_obj):
        self = MyObj()
        self.foo = db_obj['foo']
        self.bar = db_obj['bar']
        self.missing = db_obj['missing']
        self.readonly = 1
        self._context = context
        return self

    def obj_load_attr(self, attrname):
        setattr(self, attrname, 'loaded!')

    @base.remotable_classmethod
    def query(cls, context):
        obj = cls(context=context, foo=1, bar='bar')
        obj.obj_reset_changes()
        return obj

    @base.remotable
    def marco(self):
        return 'polo'

    @base.remotable
    def _update_test(self):
        self.bar = 'updated'

    @base.remotable
    def save(self):
        self.obj_reset_changes()

    @base.remotable
    def refresh(self):
        self.foo = 321
        self.bar = 'refreshed'
        self.obj_reset_changes()

    @base.remotable
    def modify_save_modify(self):
        self.bar = 'meow'
        self.save()
        self.foo = 42
        self.rel_object = MyOwnedObject(baz=42)

    def obj_make_compatible(self, primitive, target_version):
        super(MyObj, self).obj_make_compatible(primitive, target_version)
        if target_version == '1.0' and 'bar' in primitive:
            primitive['bar'] = 'old%s' % primitive['bar']


class TestObjMakeList(test.NoDBTestCase):

    def test_obj_make_list(self):
        class MyList(base.ObjectListBase, base.MasakariObject):
            fields = {
                'objects': fields.ListOfObjectsField('MyObj'),
            }

        db_objs = [{'foo': 1, 'bar': 'baz', 'missing': 'banana'},
                   {'foo': 2, 'bar': 'bat', 'missing': 'apple'},
                   ]
        mylist = base.obj_make_list('ctxt', MyList(), MyObj, db_objs)
        self.assertEqual(2, len(mylist))
        self.assertEqual('ctxt', mylist._context)
        for index, item in enumerate(mylist):
            self.assertEqual(db_objs[index]['foo'], item.foo)
            self.assertEqual(db_objs[index]['bar'], item.bar)
            self.assertEqual(db_objs[index]['missing'], item.missing)


def compare_obj(test, obj, db_obj, subs=None, allow_missing=None,
                comparators=None):
    """Compare a MasakariObject and a dict-like database object.

    This automatically converts TZ-aware datetimes and iterates over
    the fields of the object.

    :param:test: The TestCase doing the comparison
    :param:obj: The MasakariObject to examine
    :param:db_obj: The dict-like database object to use as reference
    :param:subs: A dict of objkey=dbkey field substitutions
    :param:allow_missing: A list of fields that may not be in db_obj
    :param:comparators: Map of comparator functions to use for certain fields
    """

    if subs is None:
        subs = {}
    if allow_missing is None:
        allow_missing = []
    if comparators is None:
        comparators = {}

    for key in obj.fields:
        if key in allow_missing and not obj.obj_attr_is_set(key):
            continue
        obj_val = getattr(obj, key)
        db_key = subs.get(key, key)
        db_val = db_obj[db_key]
        if isinstance(obj_val, datetime.datetime):
            obj_val = obj_val.replace(tzinfo=None)

        if key in comparators:
            comparator = comparators[key]
            comparator(db_val, obj_val)
        else:
            test.assertEqual(db_val, obj_val)


class _BaseTestCase(test.TestCase):
    def setUp(self):
        super(_BaseTestCase, self).setUp()
        self.user_id = 'fake-user'
        self.project_id = 'fake-project'
        self.context = 'masakari-context'

        base.MasakariObjectRegistry.register(MyObj)
        base.MasakariObjectRegistry.register(MyOwnedObject)

    def compare_obj(self, obj, db_obj, subs=None, allow_missing=None,
                    comparators=None):
        compare_obj(self, obj, db_obj, subs=subs, allow_missing=allow_missing,
                    comparators=comparators)

    def str_comparator(self, expected, obj_val):
        """Compare an object field to a string in the db by performing
        a simple coercion on the object field value.
        """
        self.assertEqual(expected, str(obj_val))


class _LocalTest(_BaseTestCase):
    def setUp(self):
        super(_LocalTest, self).setUp()


class _TestObject(object):
    def test_object_attrs_in_init(self):
        # Now check the test one in this file. Should be newest version
        self.assertEqual('1.6', objects.MyObj.VERSION)

    def test_hydration_type_error(self):
        primitive = {'masakari_object.name': 'MyObj',
                     'masakari_object.namespace': 'masakari',
                     'masakari_object.version': '1.5',
                     'masakari_object.data': {'foo': 'a'}}
        self.assertRaises(ValueError, MyObj.obj_from_primitive, primitive)

    def test_hydration(self):
        primitive = {'masakari_object.name': 'MyObj',
                     'masakari_object.namespace': 'masakari',
                     'masakari_object.version': '1.5',
                     'masakari_object.data': {'foo': 1}}
        real_method = MyObj._obj_from_primitive

        def _obj_from_primitive(*args):
            return real_method(*args)

        with mock.patch.object(MyObj, '_obj_from_primitive') as ofp:
            ofp.side_effect = _obj_from_primitive
            obj = MyObj.obj_from_primitive(primitive)
            ofp.assert_called_once_with(None, '1.5', primitive)
        self.assertEqual(obj.foo, 1)

    def test_hydration_version_different(self):
        primitive = {'masakari_object.name': 'MyObj',
                     'masakari_object.namespace': 'masakari',
                     'masakari_object.version': '1.2',
                     'masakari_object.data': {'foo': 1}}
        obj = MyObj.obj_from_primitive(primitive)
        self.assertEqual(obj.foo, 1)
        self.assertEqual('1.2', obj.VERSION)

    def test_hydration_bad_ns(self):
        primitive = {'masakari_object.name': 'MyObj',
                     'masakari_object.namespace': 'foo',
                     'masakari_object.version': '1.5',
                     'masakari_object.data': {'foo': 1}}
        self.assertRaises(ovo_exc.UnsupportedObjectError,
                          MyObj.obj_from_primitive, primitive)

    def test_hydration_additional_unexpected_stuff(self):
        primitive = {'masakari_object.name': 'MyObj',
                     'masakari_object.namespace': 'masakari',
                     'masakari_object.version': '1.5.1',
                     'masakari_object.data': {
                         'foo': 1,
                         'unexpected_thing': 'foobar'}}
        obj = MyObj.obj_from_primitive(primitive)
        self.assertEqual(1, obj.foo)
        self.assertFalse(hasattr(obj, 'unexpected_thing'))
        self.assertEqual('1.5.1', obj.VERSION)

    def test_dehydration(self):
        expected = {'masakari_object.name': 'MyObj',
                    'masakari_object.namespace': 'masakari',
                    'masakari_object.version': '1.6',
                    'masakari_object.data': {'foo': 1}}
        obj = MyObj(foo=1)
        obj.obj_reset_changes()
        self.assertEqual(obj.obj_to_primitive(), expected)

    def test_object_property(self):
        obj = MyObj(foo=1)
        self.assertEqual(obj.foo, 1)

    def test_object_property_type_error(self):
        obj = MyObj()

        def fail():
            obj.foo = 'a'
        self.assertRaises(ValueError, fail)

    def test_load(self):
        obj = MyObj()
        self.assertEqual(obj.bar, 'loaded!')

    def test_load_in_base(self):
        @base.MasakariObjectRegistry.register_if(False)
        class Foo(base.MasakariObject):
            fields = {'foobar': fields.IntegerField()}
        obj = Foo()
        with self.assertRaisesRegex(NotImplementedError, ".*foobar.*"):
            obj.foobar

    def test_loaded_in_primitive(self):
        obj = MyObj(foo=1)
        obj.obj_reset_changes()
        self.assertEqual(obj.bar, 'loaded!')
        expected = {'masakari_object.name': 'MyObj',
                    'masakari_object.namespace': 'masakari',
                    'masakari_object.version': '1.6',
                    'masakari_object.changes': ['bar'],
                    'masakari_object.data': {'foo': 1,
                                         'bar': 'loaded!'}}
        self.assertEqual(obj.obj_to_primitive(), expected)

    def test_changes_in_primitive(self):
        obj = MyObj(foo=123)
        self.assertEqual(obj.obj_what_changed(), set(['foo']))
        primitive = obj.obj_to_primitive()
        self.assertIn('masakari_object.changes', primitive)
        obj2 = MyObj.obj_from_primitive(primitive)
        self.assertEqual(obj2.obj_what_changed(), set(['foo']))
        obj2.obj_reset_changes()
        self.assertEqual(obj2.obj_what_changed(), set())

    def test_orphaned_object(self):
        obj = MyObj.query(self.context)
        obj._context = None
        self.assertRaises(ovo_exc.OrphanedObjectError,
                          obj._update_test)

    def test_changed_1(self):
        obj = MyObj.query(self.context)
        obj.foo = 123
        self.assertEqual(obj.obj_what_changed(), set(['foo']))
        obj._update_test()
        self.assertEqual(obj.obj_what_changed(), set(['foo', 'bar']))
        self.assertEqual(obj.foo, 123)

    def test_changed_2(self):
        obj = MyObj.query(self.context)
        obj.foo = 123
        self.assertEqual(obj.obj_what_changed(), set(['foo']))
        obj.save()
        self.assertEqual(obj.obj_what_changed(), set([]))
        self.assertEqual(obj.foo, 123)

    def test_changed_3(self):
        obj = MyObj.query(self.context)
        obj.foo = 123
        self.assertEqual(obj.obj_what_changed(), set(['foo']))
        obj.refresh()
        self.assertEqual(obj.obj_what_changed(), set([]))
        self.assertEqual(obj.foo, 321)
        self.assertEqual(obj.bar, 'refreshed')

    def test_changed_4(self):
        obj = MyObj.query(self.context)
        obj.bar = 'something'
        self.assertEqual(obj.obj_what_changed(), set(['bar']))
        obj.modify_save_modify()
        self.assertEqual(obj.obj_what_changed(), set(['foo', 'rel_object']))
        self.assertEqual(obj.foo, 42)
        self.assertEqual(obj.bar, 'meow')
        self.assertIsInstance(obj.rel_object, MyOwnedObject)

    def test_changed_with_sub_object(self):
        @base.MasakariObjectRegistry.register_if(False)
        class ParentObject(base.MasakariObject):
            fields = {'foo': fields.IntegerField(),
                      'bar': fields.ObjectField('MyObj'),
                      }
        obj = ParentObject()
        self.assertEqual(set(), obj.obj_what_changed())
        obj.foo = 1
        self.assertEqual(set(['foo']), obj.obj_what_changed())
        bar = MyObj()
        obj.bar = bar
        self.assertEqual(set(['foo', 'bar']), obj.obj_what_changed())
        obj.obj_reset_changes()
        self.assertEqual(set(), obj.obj_what_changed())
        bar.foo = 1
        self.assertEqual(set(['bar']), obj.obj_what_changed())

    def test_static_result(self):
        obj = MyObj.query(self.context)
        self.assertEqual(obj.bar, 'bar')
        result = obj.marco()
        self.assertEqual(result, 'polo')

    def test_updates(self):
        obj = MyObj.query(self.context)
        self.assertEqual(obj.foo, 1)
        obj._update_test()
        self.assertEqual(obj.bar, 'updated')

    def test_contains(self):
        obj = MyObj()
        self.assertNotIn('foo', obj)
        obj.foo = 1
        self.assertIn('foo', obj)
        self.assertNotIn('does_not_exist', obj)

    def test_obj_attr_is_set(self):
        obj = MyObj(foo=1)
        self.assertTrue(obj.obj_attr_is_set('foo'))
        self.assertFalse(obj.obj_attr_is_set('bar'))
        self.assertRaises(AttributeError, obj.obj_attr_is_set, 'bang')

    def test_obj_reset_changes_recursive(self):
        obj = MyObj(rel_object=MyOwnedObject(baz=123),
                    rel_objects=[MyOwnedObject(baz=456)])
        self.assertEqual(set(['rel_object', 'rel_objects']),
                         obj.obj_what_changed())
        obj.obj_reset_changes()
        self.assertEqual(set(['rel_object']), obj.obj_what_changed())
        self.assertEqual(set(['baz']), obj.rel_object.obj_what_changed())
        self.assertEqual(set(['baz']), obj.rel_objects[0].obj_what_changed())
        obj.obj_reset_changes(recursive=True, fields=['foo'])
        self.assertEqual(set(['rel_object']), obj.obj_what_changed())
        self.assertEqual(set(['baz']), obj.rel_object.obj_what_changed())
        self.assertEqual(set(['baz']), obj.rel_objects[0].obj_what_changed())
        obj.obj_reset_changes(recursive=True)
        self.assertEqual(set([]), obj.rel_object.obj_what_changed())
        self.assertEqual(set([]), obj.obj_what_changed())

    def test_get(self):
        obj = MyObj(foo=1)
        # Foo has value, should not get the default
        self.assertEqual(obj.get('foo', 2), 1)
        # Foo has value, should return the value without error
        self.assertEqual(obj.get('foo'), 1)
        # Bar is not loaded, so we should get the default
        self.assertEqual(obj.get('bar', 'not-loaded'), 'not-loaded')
        # Bar without a default should lazy-load
        self.assertEqual(obj.get('bar'), 'loaded!')
        # Bar now has a default, but loaded value should be returned
        self.assertEqual(obj.get('bar', 'not-loaded'), 'loaded!')
        # Invalid attribute should raise AttributeError
        self.assertRaises(AttributeError, obj.get, 'nothing')
        # ...even with a default
        self.assertRaises(AttributeError, obj.get, 'nothing', 3)

    def test_get_changes(self):
        obj = MyObj()
        self.assertEqual({}, obj.obj_get_changes())
        obj.foo = 123
        self.assertEqual({'foo': 123}, obj.obj_get_changes())
        obj.bar = 'test'
        self.assertEqual({'foo': 123, 'bar': 'test'}, obj.obj_get_changes())
        obj.obj_reset_changes()
        self.assertEqual({}, obj.obj_get_changes())

    def test_obj_fields(self):
        @base.MasakariObjectRegistry.register_if(False)
        class TestObj(base.MasakariObject):
            fields = {'foo': fields.IntegerField()}
            obj_extra_fields = ['bar']

            @property
            def bar(self):
                return 'this is bar'

        obj = TestObj()
        self.assertEqual(['foo', 'bar'], obj.obj_fields)

    def test_obj_constructor(self):
        obj = MyObj(context=self.context, foo=123, bar='abc')
        self.assertEqual(123, obj.foo)
        self.assertEqual('abc', obj.bar)
        self.assertEqual(set(['foo', 'bar']), obj.obj_what_changed())

    def test_obj_read_only(self):
        obj = MyObj(context=self.context, foo=123, bar='abc')
        obj.readonly = 1
        self.assertRaises(ovo_exc.ReadOnlyFieldError, setattr,
                          obj, 'readonly', 2)

    def test_obj_mutable_default(self):
        obj = MyObj(context=self.context, foo=123, bar='abc')
        obj.mutable_default = None
        obj.mutable_default.append('s1')
        self.assertEqual(obj.mutable_default, ['s1'])

        obj1 = MyObj(context=self.context, foo=123, bar='abc')
        obj1.mutable_default = None
        obj1.mutable_default.append('s2')
        self.assertEqual(obj1.mutable_default, ['s2'])

    def test_obj_mutable_default_set_default(self):
        obj1 = MyObj(context=self.context, foo=123, bar='abc')
        obj1.obj_set_defaults('mutable_default')
        self.assertEqual(obj1.mutable_default, [])
        obj1.mutable_default.append('s1')
        self.assertEqual(obj1.mutable_default, ['s1'])

        obj2 = MyObj(context=self.context, foo=123, bar='abc')
        obj2.obj_set_defaults('mutable_default')
        self.assertEqual(obj2.mutable_default, [])
        obj2.mutable_default.append('s2')
        self.assertEqual(obj2.mutable_default, ['s2'])

    def test_obj_repr(self):
        obj = MyObj(foo=123)
        self.assertEqual('MyObj(bar=<?>,created_at=<?>,deleted=<?>,'
                         'deleted_at=<?>,foo=123,missing=<?>,'
                         'mutable_default=<?>,readonly=<?>,rel_object=<?>,'
                         'rel_objects=<?>,updated_at=<?>)',
                         repr(obj))

    def test_obj_make_obj_compatible(self):
        subobj = MyOwnedObject(baz=1)
        subobj.VERSION = '1.2'
        obj = MyObj(rel_object=subobj)
        obj.obj_relationships = {
            'rel_object': [('1.5', '1.1'), ('1.7', '1.2')],
        }
        orig_primitive = obj.obj_to_primitive()['masakari_object.data']
        with mock.patch.object(subobj, 'obj_make_compatible') as mock_compat:
            primitive = copy.deepcopy(orig_primitive)
            obj._obj_make_obj_compatible(primitive, '1.8', 'rel_object')
            self.assertFalse(mock_compat.called)

        with mock.patch.object(subobj, 'obj_make_compatible') as mock_compat:
            primitive = copy.deepcopy(orig_primitive)
            obj._obj_make_obj_compatible(primitive, '1.7', 'rel_object')
            mock_compat.assert_called_once_with(
                primitive['rel_object']['masakari_object.data'], '1.2')

        with mock.patch.object(subobj, 'obj_make_compatible') as mock_compat:
            primitive = copy.deepcopy(orig_primitive)
            obj._obj_make_obj_compatible(primitive, '1.6', 'rel_object')
            mock_compat.assert_called_once_with(
                primitive['rel_object']['masakari_object.data'], '1.1')
            self.assertEqual('1.1', primitive[
                'rel_object']['masakari_object.version'])

        with mock.patch.object(subobj, 'obj_make_compatible') as mock_compat:
            primitive = copy.deepcopy(orig_primitive)
            obj._obj_make_obj_compatible(primitive, '1.5', 'rel_object')
            mock_compat.assert_called_once_with(
                primitive['rel_object']['masakari_object.data'], '1.1')
            self.assertEqual('1.1', primitive[
                'rel_object']['masakari_object.version'])

        with mock.patch.object(subobj, 'obj_make_compatible') as mock_compat:
            primitive = copy.deepcopy(orig_primitive)
            obj._obj_make_obj_compatible(primitive, '1.4', 'rel_object')
            self.assertFalse(mock_compat.called)
            self.assertNotIn('rel_object', primitive)

    def test_obj_make_compatible_hits_sub_objects(self):
        subobj = MyOwnedObject(baz=1)
        obj = MyObj(foo=123, rel_object=subobj)
        obj.obj_relationships = {'rel_object': [('1.0', '1.0')]}
        with mock.patch.object(obj, '_obj_make_obj_compatible') as mock_compat:
            obj.obj_make_compatible({'rel_object': 'foo'}, '1.10')
            mock_compat.assert_called_once_with({'rel_object': 'foo'}, '1.10',
                                                'rel_object')

    def test_obj_make_compatible_skips_unset_sub_objects(self):
        obj = MyObj(foo=123)
        obj.obj_relationships = {'rel_object': [('1.0', '1.0')]}
        with mock.patch.object(obj, '_obj_make_obj_compatible') as mock_compat:
            obj.obj_make_compatible({'rel_object': 'foo'}, '1.10')
            self.assertFalse(mock_compat.called)

    def test_obj_make_compatible_doesnt_skip_falsey_sub_objects(self):
        @base.MasakariObjectRegistry.register_if(False)
        class MyList(base.ObjectListBase, base.MasakariObject):
            VERSION = '1.2'
            fields = {'objects': fields.ListOfObjectsField('MyObjElement')}
            obj_relationships = {
                'objects': [('1.1', '1.1'), ('1.2', '1.2')],
            }

        mylist = MyList(objects=[])

        @base.MasakariObjectRegistry.register_if(False)
        class MyOwner(base.MasakariObject):
            VERSION = '1.2'
            fields = {'mylist': fields.ObjectField('MyList')}
            obj_relationships = {
                'mylist': [('1.1', '1.1')],
            }

        myowner = MyOwner(mylist=mylist)
        primitive = myowner.obj_to_primitive('1.1')
        self.assertIn('mylist', primitive['masakari_object.data'])

    def test_obj_make_compatible_handles_list_of_objects(self):
        subobj = MyOwnedObject(baz=1)
        obj = MyObj(rel_objects=[subobj])
        obj.obj_relationships = {'rel_objects': [('1.0', '1.123')]}

        def fake_make_compat(primitive, version):
            self.assertEqual('1.123', version)
            self.assertIn('baz', primitive)

        with mock.patch.object(subobj, 'obj_make_compatible') as mock_mc:
            mock_mc.side_effect = fake_make_compat
            obj.obj_to_primitive('1.0')
            self.assertTrue(mock_mc.called)

    def test_delattr(self):
        obj = MyObj(bar='foo')
        del obj.bar

        # Should appear unset now
        self.assertFalse(obj.obj_attr_is_set('bar'))

        # Make sure post-delete, references trigger lazy loads
        self.assertEqual('loaded!', getattr(obj, 'bar'))

    def test_delattr_unset(self):
        obj = MyObj()
        self.assertRaises(AttributeError, delattr, obj, 'bar')


class TestObject(_LocalTest, _TestObject):
    def test_set_defaults(self):
        obj = MyObj()
        obj.obj_set_defaults('foo')
        self.assertTrue(obj.obj_attr_is_set('foo'))
        self.assertEqual(1, obj.foo)

    def test_set_defaults_no_default(self):
        obj = MyObj()
        self.assertRaises(ovo_exc.ObjectActionError,
                          obj.obj_set_defaults, 'bar')

    def test_set_all_defaults(self):
        obj = MyObj()
        obj.obj_set_defaults()
        self.assertEqual(set(['deleted', 'foo', 'mutable_default']),
                         obj.obj_what_changed())
        self.assertEqual(1, obj.foo)

    def test_set_defaults_not_overwrite(self):
        obj = MyObj(deleted=True)
        obj.obj_set_defaults()
        self.assertEqual(1, obj.foo)
        self.assertTrue(obj.deleted)


class TestRegistry(test.NoDBTestCase):
    @mock.patch('masakari.objects.base.objects')
    def test_hook_chooses_newer_properly(self, mock_objects):
        reg = base.MasakariObjectRegistry()
        reg.registration_hook(MyObj, 0)

        class MyNewerObj(object):
            VERSION = '1.123'

            @classmethod
            def obj_name(cls):
                return 'MyObj'

        self.assertEqual(MyObj, mock_objects.MyObj)
        reg.registration_hook(MyNewerObj, 0)
        self.assertEqual(MyNewerObj, mock_objects.MyObj)

    @mock.patch('masakari.objects.base.objects')
    def test_hook_keeps_newer_properly(self, mock_objects):
        reg = base.MasakariObjectRegistry()
        reg.registration_hook(MyObj, 0)

        class MyOlderObj(object):
            VERSION = '1.1'

            @classmethod
            def obj_name(cls):
                return 'MyObj'

        self.assertEqual(MyObj, mock_objects.MyObj)
        reg.registration_hook(MyOlderObj, 0)
        self.assertEqual(MyObj, mock_objects.MyObj)


# NOTE(Dinesh_Bhor): The hashes in this list should only be changed if
# they come with a corresponding version bump in the affected
# objects
object_data = {
    'FailoverSegment': '1.1-9cecc07c111f647b32d560f19f1f5db9',
    'FailoverSegmentList': '1.0-dfc5c6f5704d24dcaa37b0bbb03cbe60',
    'Host': '1.2-f05735b156b687bc916d46b551bc45e3',
    'HostList': '1.0-25ebe1b17fbd9f114fae8b6a10d198c0',
    'Notification': '1.1-91e3a051078e35300e325a3e2ae5fde5',
    'NotificationProgressDetails': '1.0-fc611ac932b719fbc154dbe34bb8edee',
    'NotificationList': '1.0-25ebe1b17fbd9f114fae8b6a10d198c0',
    'EventType': '1.0-d1d2010a7391fa109f0868d964152607',
    'ExceptionNotification': '1.0-1187e93f564c5cca692db76a66cda2a6',
    'ExceptionPayload': '1.0-96f178a12691e3ef0d8e3188fc481b90',
    'HostApiNotification': '1.0-1187e93f564c5cca692db76a66cda2a6',
    'HostApiPayload': '1.0-20603e23a06477975b860459ba684d34',
    'HostApiPayloadBase': '1.1-e02e80de08d0b584ee7a0a0320bfd029',
    'NotificationApiPayload': '1.0-c050869a1f4aed23e7645bd4d1830ecd',
    'NotificationApiPayloadBase': '1.0-cda8d53a77e64f83e3782fc9c4d499bb',
    'NotificationApiNotification': '1.0-1187e93f564c5cca692db76a66cda2a6',
    'NotificationPublisher': '1.0-bbbc1402fb0e443a3eb227cc52b61545',
    'MyObj': '1.6-ee7b607402fbfb3390a92ab7199e0d88',
    'MyOwnedObject': '1.0-fec853730bd02d54cc32771dd67f08a0',
    'SegmentApiNotification': '1.0-1187e93f564c5cca692db76a66cda2a6',
    'SegmentApiPayload': '1.1-e34e1c772e16e9ad492067ee98607b1d',
    'SegmentApiPayloadBase': '1.1-6a1db76f3e825f92196fc1a11508d886'
}


def get_masakari_objects():
    """Get masakari versioned objects

    This returns a dict of versioned objects which are
    in the Masakari project namespace only. ie excludes
    objects from os-vif and other 3rd party modules

    :return: a dict mapping class names to lists of versioned objects
    """

    all_classes = base.MasakariObjectRegistry.obj_classes()
    masakari_classes = {}
    for name in all_classes:
        objclasses = all_classes[name]
        if (objclasses[0].OBJ_PROJECT_NAMESPACE != (
                base.MasakariObject.OBJ_PROJECT_NAMESPACE)):
            continue
        masakari_classes[name] = objclasses
    return masakari_classes


class TestObjectVersions(test.NoDBTestCase, _BaseTestCase):
    def setUp(self):
        super(test.NoDBTestCase, self).setUp()
        base.MasakariObjectRegistry.register_notification_objects()

    def test_versions(self):
        checker = fixture.ObjectVersionChecker(
            get_masakari_objects())
        fingerprints = checker.get_hashes()

        if os.getenv('GENERATE_HASHES'):
            open('object_hashes.txt', 'w').write(
                pprint.pformat(fingerprints))
            raise test.TestingException(
                'Generated hashes in object_hashes.txt')

        expected, actual = checker.test_hashes(object_data)
        self.assertEqual(expected, actual,
                         'Some objects have changed; please make sure the '
                         'versions have been bumped, and then update their '
                         'hashes here.')

    def test_obj_make_compatible(self):
        base.MasakariObjectRegistry.register(segment.FailoverSegment)

        # Iterate all object classes and verify that we can run
        # obj_make_compatible with every older version than current.
        # This doesn't actually test the data conversions, but it at least
        # makes sure the method doesn't blow up on something basic like
        # expecting the wrong version format.

        # Hold a dictionary of args/kwargs that need to get passed into
        # __init__() for specific classes. The key in the dictionary is
        # the obj_class that needs the init args/kwargs.
        init_args = fake_args.init_args
        init_kwargs = fake_args.init_kwargs

        checker = fixture.ObjectVersionChecker(
            base.MasakariObjectRegistry.obj_classes())
        checker.test_compatibility_routines(use_manifest=True,
                                            init_args=init_args,
                                            init_kwargs=init_kwargs)

    def test_list_obj_make_compatible(self):
        @base.MasakariObjectRegistry.register_if(False)
        class TestObj(base.MasakariObject):
            VERSION = '1.4'
            fields = {'foo': fields.IntegerField()}

        @base.MasakariObjectRegistry.register_if(False)
        class TestListObj(base.ObjectListBase, base.MasakariObject):
            VERSION = '1.5'
            fields = {'objects': fields.ListOfObjectsField('TestObj')}
            obj_relationships = {
                'objects': [('1.0', '1.1'), ('1.1', '1.2'),
                            ('1.3', '1.3'), ('1.5', '1.4')]
            }

        my_list = TestListObj()
        my_obj = TestObj(foo=1)
        my_list.objects = [my_obj]
        primitive = my_list.obj_to_primitive(target_version='1.5')
        primitive_data = primitive['masakari_object.data']
        obj_primitive = my_obj.obj_to_primitive(target_version='1.4')
        obj_primitive_data = obj_primitive['masakari_object.data']
        with mock.patch.object(TestObj, 'obj_make_compatible') as comp:
            my_list.obj_make_compatible(primitive_data, '1.1')
            comp.assert_called_with(obj_primitive_data,
                                    '1.2')

    def test_list_obj_make_compatible_when_no_objects(self):
        # Test to make sure obj_make_compatible works with no 'objects'
        # If a List object ever has a version that did not contain the
        # 'objects' key, we need to make sure converting back to that version
        # doesn't cause backporting problems.
        @base.MasakariObjectRegistry.register_if(False)
        class TestObj(base.MasakariObject):
            VERSION = '1.1'
            fields = {'foo': fields.IntegerField()}

        @base.MasakariObjectRegistry.register_if(False)
        class TestListObj(base.ObjectListBase, base.MasakariObject):
            VERSION = '1.1'
            fields = {'objects': fields.ListOfObjectsField('TestObj')}
            obj_relationships = {
                'objects': [('1.1', '1.1')]
            }

        my_list = TestListObj()
        my_list.objects = [TestObj(foo=1)]
        primitive = my_list.obj_to_primitive(target_version='1.1')
        primitive_data = primitive['masakari_object.data']
        my_list.obj_make_compatible(primitive_data,
                                    target_version='1.0')
        self.assertNotIn('objects', primitive_data,
                         "List was backported to before 'objects' existed."
                         " 'objects' should not be in the primitive.")


class TestObjEqualPrims(_BaseTestCase):

    def test_object_equal(self):
        obj1 = MyObj(foo=1, bar='goodbye')
        obj1.obj_reset_changes()
        obj2 = MyObj(foo=1, bar='goodbye')
        obj2.obj_reset_changes()
        obj2.bar = 'goodbye'
        # obj2 will be marked with field 'three' updated
        self.assertTrue(base.obj_equal_prims(obj1, obj2),
                        "Objects that differ only because one a is marked "
                        "as updated should be equal")

    def test_object_not_equal(self):
        obj1 = MyObj(foo=1, bar='goodbye')
        obj1.obj_reset_changes()
        obj2 = MyObj(foo=1, bar='hello')
        obj2.obj_reset_changes()
        self.assertFalse(base.obj_equal_prims(obj1, obj2),
                         "Objects that differ in any field "
                         "should not be equal")

    def test_object_ignore_equal(self):
        obj1 = MyObj(foo=1, bar='goodbye')
        obj1.obj_reset_changes()
        obj2 = MyObj(foo=1, bar='hello')
        obj2.obj_reset_changes()
        self.assertTrue(base.obj_equal_prims(obj1, obj2, ['bar']),
                        "Objects that only differ in an ignored field "
                        "should be equal")


class TestObjMethodOverrides(test.NoDBTestCase):
    def test_obj_reset_changes(self):
        args = inspect.getargspec(base.MasakariObject.obj_reset_changes)
        obj_classes = base.MasakariObjectRegistry.obj_classes()
        for obj_name in obj_classes:
            obj_class = obj_classes[obj_name][0]
            self.assertEqual(args,
                    inspect.getargspec(obj_class.obj_reset_changes))



# Copyright (c) 2016 NTT Data
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


api_opts = [
    cfg.ListOpt("extensions_blacklist",
                default=[],
                deprecated_for_removal=True,
                deprecated_group="osapi_v1",
                help="""
*DEPRECATED*

This option is a list of all of the v2.1 API extensions to never load. However,
it will be removed in the near future, after which the all the functionality
that was previously in extensions will be part of the standard API, and thus
always accessible.

* Possible values:

    A list of strings, each being the alias of an extension that you do not
    wish to load.

* Services that use this:

    ``masakari-api``

* Related options:

    enabled, extensions_whitelist
"""),
    cfg.ListOpt("extensions_whitelist",
                default=[],
                deprecated_for_removal=True,
                deprecated_group="osapi_v1",
                help="""
*DEPRECATED*

This is a list of extensions. If it is empty, then *all* extensions except
those specified in the extensions_blacklist option will be loaded. If it is not
empty, then only those extensions in this list will be loaded, provided that
they are also not in the extensions_blacklist option. Once this deprecated
option is removed, after which the all the functionality that was previously in
extensions will be part of the standard API, and thus always accessible.

* Possible values:

    A list of strings, each being the alias of an extension that you wish to
    load, or an empty list, which indicates that all extensions are to be run.

* Services that use this:

    ``masakari-api``

* Related options:

    enabled, extensions_blacklist
"""),
    cfg.StrOpt("project_id_regex",
               default=None,
               deprecated_for_removal=True,
               deprecated_group="osapi_v1",
               help="""
*DEPRECATED*

This option is a string representing a regular expression (regex) that matches
the project_id as contained in URLs. If not set, it will match normal UUIDs
created by keystone.

* Possible values:

    A string representing any legal regular expression

* Services that use this:

    ``masakari-api``

* Related options:

    None
"""),
]

api_opts_group = cfg.OptGroup(name="osapi_v1", title="API v1 Options")


def register_opts(conf):
    conf.register_group(api_opts_group)
    conf.register_opts(api_opts, api_opts_group)


def list_opts():
    return {api_opts_group: api_opts}



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Masakari base exception handling.

Includes decorator for re-raising Masakari-type exceptions.

SHOULD include dedicated exception logging.

"""

import functools
from http import client as http
import inspect
import sys

from oslo_log import log as logging
from oslo_utils import excutils
import webob.exc
from webob import util as woutil

import masakari.conf
from masakari.i18n import _
from masakari import safe_utils
from masakari import utils

LOG = logging.getLogger(__name__)


CONF = masakari.conf.CONF


class ConvertedException(webob.exc.WSGIHTTPException):
    def __init__(self, code, title="", explanation=""):
        self.code = int(code)
        # There is a strict rule about constructing status line for HTTP:
        # '...Status-Line, consisting of the protocol version followed by a
        # numeric status code and its associated textual phrase, with each
        # element separated by SP characters'
        # (http://www.faqs.org/rfcs/rfc2616.html)
        # 'code' and 'title' can not be empty because they correspond
        # to numeric status code and its associated text
        if title:
            self.title = title
        else:
            try:
                self.title = woutil.status_reasons[self.code]
            except KeyError:
                msg = "Improper or unknown HTTP status code used: %d"
                LOG.error(msg, code)
                self.title = woutil.status_generic_reasons[self.code // 100]
        self.explanation = explanation
        super(ConvertedException, self).__init__()


def _cleanse_dict(original):
    """Strip all admin_password, new_pass, rescue_pass keys from a dict."""
    return {k: v for k, v in original.items() if "_pass" not in k}


def wrap_exception(notifier=None, get_notifier=None):
    """This decorator wraps a method to catch any exceptions that may
    get thrown. It also optionally sends the exception to the notification
    system.
    """
    def inner(f):
        def wrapped(self, context, *args, **kw):
            # Don't store self or context in the payload, it now seems to
            # contain confidential information.
            try:
                return f(self, context, *args, **kw)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    if notifier or get_notifier:
                        payload = dict(exception=e)
                        wrapped_func = safe_utils.get_wrapped_function(f)
                        call_dict = inspect.getcallargs(wrapped_func, self,
                                                        context, *args, **kw)
                        # self can't be serialized and shouldn't be in the
                        # payload
                        call_dict.pop('self', None)
                        cleansed = _cleanse_dict(call_dict)
                        payload.update({'args': cleansed})

                        # If f has multiple decorators, they must use
                        # functools.wraps to ensure the name is
                        # propagated.
                        event_type = f.__name__

                        (notifier or get_notifier()).error(context,
                                                           event_type,
                                                           payload)

        return functools.wraps(f)(wrapped)
    return inner


class MasakariException(Exception):
    """Base Masakari Exception

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.

    """
    msg_fmt = _("An unknown exception occurred.")
    code = http.INTERNAL_SERVER_ERROR
    headers = {}
    safe = False

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            try:
                message = self.msg_fmt % kwargs

            except Exception:
                exc_info = sys.exc_info()
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in kwargs.items():
                    LOG.error("%s: %s" % (name, value))    # noqa

                if CONF.fatal_exception_format_errors:
                    utils.reraise(*exc_info)
                else:
                    # at least get the core message out if something happened
                    message = self.msg_fmt

        self.message = message
        super(MasakariException, self).__init__(message)

    def format_message(self):
        # NOTE: use the first argument to the python Exception object
        # which should be our full MasakariException message, (see __init__)
        return self.args[0]


class APIException(MasakariException):
    msg_fmt = _("Error while requesting %(service)s API.")

    def __init__(self, message=None, **kwargs):
        if 'service' not in kwargs:
            kwargs['service'] = 'unknown'
        super(APIException, self).__init__(message, **kwargs)


class APITimeout(APIException):
    msg_fmt = _("Timeout while requesting %(service)s API.")


class Conflict(MasakariException):
    msg_fmt = _("Conflict")
    code = http.CONFLICT


class Invalid(MasakariException):
    msg_fmt = _("Bad Request - Invalid Parameters")
    code = http.BAD_REQUEST


class InvalidName(Invalid):
    msg_fmt = _("An invalid 'name' value was provided. "
                "The name must be: %(reason)s")


class InvalidInput(Invalid):
    msg_fmt = _("Invalid input received: %(reason)s")


class InvalidAPIVersionString(Invalid):
    msg_fmt = _("API Version String %(version)s is of invalid format. Must "
                "be of format MajorNum.MinorNum.")


class MalformedRequestBody(MasakariException):
    msg_fmt = _("Malformed message body: %(reason)s")


# NOTE: NotFound should only be used when a 404 error is
# appropriate to be returned
class NotFound(MasakariException):
    msg_fmt = _("Resource could not be found.")
    code = http.NOT_FOUND


class ConfigNotFound(NotFound):
    msg_fmt = _("Could not find config at %(path)s")


class Forbidden(MasakariException):
    msg_fmt = _("Forbidden")
    code = http.FORBIDDEN


class AdminRequired(Forbidden):
    msg_fmt = _("User does not have admin privileges")


class PolicyNotAuthorized(Forbidden):
    msg_fmt = _("Policy doesn't allow %(action)s to be performed.")


class PasteAppNotFound(MasakariException):
    msg_fmt = _("Could not load paste app '%(name)s' from %(path)s")


class InvalidContentType(Invalid):
    msg_fmt = _("Invalid content type %(content_type)s.")


class VersionNotFoundForAPIMethod(Invalid):
    msg_fmt = _("API version %(version)s is not supported on this method.")


class InvalidGlobalAPIVersion(Invalid):
    msg_fmt = _("Version %(req_ver)s is not supported by the API. Minimum "
                "is %(min_ver)s and maximum is %(max_ver)s.")


class ApiVersionsIntersect(Invalid):
    msg_fmt = _("Version of %(name) %(min_ver) %(max_ver) intersects "
                "with another versions.")


class ValidationError(Invalid):
    msg_fmt = "%(detail)s"


class InvalidSortKey(Invalid):
    msg_fmt = _("Sort key supplied was not valid.")


class MarkerNotFound(NotFound):
    msg_fmt = _("Marker %(marker)s could not be found.")


class FailoverSegmentNotFound(NotFound):
    msg_fmt = _("No failover segment with id %(id)s.")


class HostNotFound(NotFound):
    msg_fmt = _("No host with id %(id)s.")


class NotificationNotFound(NotFound):
    msg_fmt = _("No notification with id %(id)s.")


class FailoverSegmentNotFoundByName(FailoverSegmentNotFound):
    msg_fmt = _("Failover segment with name %(segment_name)s could not "
                "be found.")


class HostNotFoundByName(HostNotFound):
    msg_fmt = _("Host with name %(host_name)s could not be found.")


class ComputeNotFoundByName(NotFound):
    msg_fmt = _("Compute service with name %(compute_name)s could not "
                "be found.")


class FailoverSegmentExists(MasakariException):
    msg_fmt = _("Failover segment with name %(name)s already exists.")


class HostExists(MasakariException):
    msg_fmt = _("Host with name %(name)s already exists.")


class Unauthorized(MasakariException):
    msg_fmt = _("Not authorized.")
    code = http.UNAUTHORIZED


class ObjectActionError(MasakariException):
    msg_fmt = _('Object action %(action)s failed because: %(reason)s')


class OrphanedObjectError(MasakariException):
    msg_fmt = _('Cannot call %(method)s on orphaned %(objtype)s object')


class DuplicateNotification(Invalid):
    msg_fmt = _('Duplicate notification received for type: %(type)s')
    code = http.CONFLICT


class HostOnMaintenanceError(Invalid):
    msg_fmt = _('Host %(host_name)s is already under maintenance.')
    code = http.CONFLICT


class HostRecoveryFailureException(MasakariException):
    msg_fmt = _('Failed to execute host recovery.')


class InstanceRecoveryFailureException(MasakariException):
    msg_fmt = _('Failed to execute instance recovery workflow.')


class SkipInstanceRecoveryException(MasakariException):
    msg_fmt = _('Skipping execution of instance recovery workflow.')


class SkipProcessRecoveryException(MasakariException):
    msg_fmt = _('Skipping execution of process recovery workflow.')


class SkipHostRecoveryException(MasakariException):
    msg_fmt = _('Skipping execution of host recovery workflow.')


class ProcessRecoveryFailureException(MasakariException):
    msg_fmt = _('Failed to execute process recovery workflow.')


class DBNotAllowed(MasakariException):
    msg_fmt = _('%(binary)s attempted direct database access which is '
                'not allowed by policy')


class FailoverSegmentInUse(Conflict):
    msg_fmt = _("Failover segment %(uuid)s can't be updated as it is in-use "
                "to process notifications.")


class HostInUse(Conflict):
    msg_fmt = _("Host %(uuid)s can't be updated as it is in-use to process "
                "notifications.")


class ReservedHostsUnavailable(MasakariException):
    msg_fmt = _('No reserved_hosts available for evacuation.')


class LockAlreadyAcquired(MasakariException):
    msg_fmt = _('Lock is already acquired on %(resource)s.')


class IgnoreInstanceRecoveryException(MasakariException):
    msg_fmt = _('Instance recovery is ignored.')


class HostNotFoundUnderFailoverSegment(HostNotFound):
    msg_fmt = _("Host '%(host_uuid)s' under failover_segment "
                "'%(segment_uuid)s' could not be found.")


class InstanceEvacuateFailed(MasakariException):
    msg_fmt = _("Failed to evacuate instance %(instance_uuid)s")


class FailoverSegmentDisabled(MasakariException):
    msg_fmt = _('Failover segment is disabled.')



# Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_log import log as logging
import paste.urlmap
from urllib import request

from masakari.api.openstack import wsgi


LOG = logging.getLogger(__name__)


_quoted_string_re = r'"[^"\\]*(?:\\.[^"\\]*)*"'
_option_header_piece_re = re.compile(
    r';\s*([^\s;=]+|%s)\s*'
    r'(?:=\s*([^;]+|%s))?\s*' % (_quoted_string_re, _quoted_string_re))


def unquote_header_value(value):
    """Unquotes a header value.
    This does not use the real unquoting but what browsers are actually
    using for quoting.

    :param value: the header value to unquote.
    """
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]
    return value


def parse_list_header(value):
    """Parse lists as described by RFC 2068 Section 2.

    In particular, parse comma-separated lists where the elements of
    the list may include quoted-strings.  A quoted-string could
    contain a comma.  A non-quoted string could have quotes in the
    middle.  Quotes are removed automatically after parsing.

    The return value is a standard :class:`list`:

    >>> parse_list_header('token, "quoted value"')
    ['token', 'quoted value']

    :param value: a string with a list header.
    :return: :class:`list`
    """
    result = []
    for item in request.parse_http_list(value):
        if item[:1] == item[-1:] == '"':
            item = unquote_header_value(item[1:-1])
        result.append(item)
    return result


def parse_options_header(value):
    """Parse a ``Content-Type`` like header into a tuple with the content
    type and the options:

    >>> parse_options_header('Content-Type: text/html; mimetype=text/html')
    ('Content-Type:', {'mimetype': 'text/html'})

    :param value: the header to parse.
    :return: (str, options)
    """
    def _tokenize(string):
        for match in _option_header_piece_re.finditer(string):
            key, value = match.groups()
            key = unquote_header_value(key)
            if value is not None:
                value = unquote_header_value(value)
            yield key, value

    if not value:
        return '', {}

    parts = _tokenize(';' + value)
    name = next(parts)[0]
    extra = dict(parts)
    return name, extra


class Accept(object):
    def __init__(self, value):
        self._content_types = [parse_options_header(v) for v in
                               parse_list_header(value)]

    def best_match(self, supported_content_types):
        best_quality = -1
        best_content_type = None
        best_params = {}
        best_match = '*/*'

        for content_type in supported_content_types:
            for content_mask, params in self._content_types:
                try:
                    quality = float(params.get('q', 1))
                except ValueError:
                    continue

                if quality < best_quality:
                    continue
                elif best_quality == quality:
                    if best_match.count('*') <= content_mask.count('*'):
                        continue

                if self._match_mask(content_mask, content_type):
                    best_quality = quality
                    best_content_type = content_type
                    best_params = params
                    best_match = content_mask

        return best_content_type, best_params

    def _match_mask(self, mask, content_type):
        if '*' not in mask:
            return content_type == mask
        if mask == '*/*':
            return True
        mask_major = mask[:-2]
        content_type_major = content_type.split('/', 1)[0]
        return content_type_major == mask_major


def urlmap_factory(loader, global_conf, **local_conf):
    if 'not_found_app' in local_conf:
        not_found_app = local_conf.pop('not_found_app')
    else:
        not_found_app = global_conf.get('not_found_app')
    if not_found_app:
        not_found_app = loader.get_app(not_found_app, global_conf=global_conf)
    urlmap = URLMap(not_found_app=not_found_app)
    for path, app_name in local_conf.items():
        path = paste.urlmap.parse_path_expression(path)
        app = loader.get_app(app_name, global_conf=global_conf)
        urlmap[path] = app
    return urlmap


class URLMap(paste.urlmap.URLMap):
    def _match(self, host, port, path_info):
        """Find longest match for a given URL path."""
        for (domain, app_url), app in self.applications:
            if domain and domain != host and domain != host + ':' + port:
                continue
            if (path_info == app_url or
                    path_info.startswith(app_url + '/')):
                return app, app_url

        return None, None

    def _set_script_name(self, app, app_url):
        def wrap(environ, start_response):
            environ['SCRIPT_NAME'] += app_url
            return app(environ, start_response)

        return wrap

    def _munge_path(self, app, path_info, app_url):
        def wrap(environ, start_response):
            environ['SCRIPT_NAME'] += app_url
            environ['PATH_INFO'] = path_info[len(app_url):]
            return app(environ, start_response)

        return wrap

    def _path_strategy(self, host, port, path_info):
        """Check path suffix for MIME type and path prefix for API version."""
        mime_type = app = app_url = None

        parts = path_info.rsplit('.', 1)
        if len(parts) > 1:
            possible_type = 'application/' + parts[1]
            if possible_type in wsgi.get_supported_content_types():
                mime_type = possible_type

        parts = path_info.split('/')
        if len(parts) > 1:
            possible_app, possible_app_url = self._match(host, port, path_info)
            # Don't use prefix if it ends up matching default
            if possible_app and possible_app_url:
                app_url = possible_app_url
                app = self._munge_path(possible_app, path_info, app_url)

        return mime_type, app, app_url

    def _content_type_strategy(self, host, port, environ):
        """Check Content-Type header for API version."""
        app = None
        params = parse_options_header(environ.get('CONTENT_TYPE', ''))[1]
        if 'version' in params:
            app, app_url = self._match(host, port, '/v' + params['version'])
            if app:
                app = self._set_script_name(app, app_url)

        return app

    def _accept_strategy(self, host, port, environ, supported_content_types):
        """Check Accept header for best matching MIME type and API version."""
        accept = Accept(environ.get('HTTP_ACCEPT', ''))

        app = None

        # Find the best match in the Accept header
        mime_type, params = accept.best_match(supported_content_types)
        if 'version' in params:
            app, app_url = self._match(host, port, '/v' + params['version'])
            if app:
                app = self._set_script_name(app, app_url)

        return mime_type, app

    def __call__(self, environ, start_response):
        host = environ.get('HTTP_HOST', environ.get('SERVER_NAME')).lower()
        if ':' in host:
            host, port = host.split(':', 1)
        else:
            if environ['wsgi.url_scheme'] == 'http':
                port = '80'
            else:
                port = '443'

        path_info = environ['PATH_INFO']
        path_info = self.normalize_url(path_info, False)[1]

        # The MIME type for the response is determined in one of two ways:
        # 1) URL path suffix (eg /servers/detail.json)
        # 2) Accept header (eg application/json;q=0.8, application/xml;q=0.2)

        # The API version is determined in one of three ways:
        # 1) URL path prefix (eg /v1.1/tenant/servers/detail)
        # 2) Content-Type header (eg application/json;version=1.1)
        # 3) Accept header (eg application/json;q=0.8;version=1.1)

        supported_content_types = list(wsgi.get_supported_content_types())

        mime_type, app, app_url = self._path_strategy(host, port, path_info)

        # Accept application/atom+xml for the index query of each API
        # version mount point as well as the root index
        if (app_url and app_url + '/' == path_info) or path_info == '/':
            supported_content_types.append('application/atom+xml')

        if not app:
            app = self._content_type_strategy(host, port, environ)

        if not mime_type or not app:
            possible_mime_type, possible_app = self._accept_strategy(
                host, port, environ, supported_content_types)
            if possible_mime_type and not mime_type:
                mime_type = possible_mime_type
            if possible_app and not app:
                app = possible_app

        if not mime_type:
            mime_type = 'application/json'

        if not app:
            # Didn't match a particular version, probably matches default
            app, app_url = self._match(host, port, path_info)
            if app:
                app = self._munge_path(app, path_info, app_url)

        if app:
            environ['masakari.best_content_type'] = mime_type
            return app(environ, start_response)

        LOG.debug('Could not find application for %s', environ['PATH_INFO'])
        environ['paste.urlmap_object'] = self
        return self.not_found_application(environ, start_response)



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the hosts api."""

from http import client as http
from unittest import mock

import ddt
from oslo_serialization import jsonutils
from webob import exc

from masakari.api.openstack.ha import hosts
from masakari import exception
from masakari.ha import api as ha_api
from masakari.objects import base as obj_base
from masakari.objects import host as host_obj
from masakari.objects import segment as segment_obj
from masakari import test
from masakari.tests.unit.api.openstack import fakes
from masakari.tests.unit import fakes as fakes_data
from masakari.tests import uuidsentinel


def _make_host_obj(host_dict):
    return host_obj.Host(**host_dict)


def _make_hosts_list(hosts_list):
    return host_obj.Host(objects=[
        _make_host_obj(a) for a in hosts_list])


@ddt.ddt
class HostTestCase(test.TestCase):
    """Test Case for host api."""

    bad_request = exception.ValidationError

    def _set_up(self):
        self.controller = hosts.HostsController()
        self.req = fakes.HTTPRequest.blank(
            '/v1/segments/%s/hosts' % uuidsentinel.fake_segment1,
            use_admin_context=True)
        self.context = self.req.environ['masakari.context']

    def setUp(self):
        super(HostTestCase, self).setUp()
        self._set_up()
        self.failover_segment = fakes_data.create_fake_failover_segment(
            name="segment1", id=1, description="failover_segment for compute",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )
        self.host = fakes_data.create_fake_host(
            name="host_1", id=1, reserved=False, on_maintenance=False,
            type="fake", control_attributes="fake-control_attributes",
            uuid=uuidsentinel.fake_host_1,
            failover_segment=self.failover_segment
        )
        self.host_2 = fakes_data.create_fake_host(
            name="host_2", id=2, reserved=False, on_maintenance=False,
            type="fake", control_attributes="fake-control_attributes",
            uuid=uuidsentinel.fake_host_2,
            failover_segment=self.failover_segment
        )
        self.host_list = [self.host, self.host_2]
        self.host_list_obj = _make_hosts_list(self.host_list)

    @property
    def app(self):
        return fakes.wsgi_app_v1(init_only='os-hosts')

    def _assert_host_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The host objects were not equal")

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    @mock.patch.object(ha_api.HostAPI, 'get_all')
    def test_index(self, mock_get_all, mock_segment):
        mock_segment.return_value = mock.Mock()
        mock_get_all.return_value = self.host_list

        result = self.controller.index(self.req, uuidsentinel.fake_segment1)
        result = result['hosts']
        self._assert_host_data(self.host_list_obj, _make_hosts_list(result))

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    @mock.patch.object(ha_api.HostAPI, 'get_all')
    def test_index_valid_on_maintenance(self, mock_get_all, mock_segment):
        mock_segment.return_value = mock.Mock()

        self.host_list[0]['on_maintenance'] = True
        self.host_list[1]['on_maintenance'] = True
        mock_get_all.return_value = self.host_list
        for parameter in ['1', 't', 'true', 'on', 'y', 'yes']:
            req = fakes.HTTPRequest.blank(
                '/v1/segments/%s/hosts?on_maintenance=''%s' % (
                    uuidsentinel.fake_segment1, parameter),
                use_admin_context=True)
            result = self.controller.index(req, uuidsentinel.fake_segment1)
            self.assertIn('hosts', result)
            for host in result['hosts']:
                self.assertTrue(host['on_maintenance'])

        self.host_list[0]['on_maintenance'] = False
        self.host_list[1]['on_maintenance'] = False
        mock_get_all.return_value = self.host_list
        for parameter in ['0', 'f', 'false', 'off', 'n', 'no']:
            req = fakes.HTTPRequest.blank(
                '/v1/segments/%s/hosts?on_maintenance=''%s' % (
                    uuidsentinel.fake_segment1, parameter),
                use_admin_context=True)
            result = self.controller.index(req, uuidsentinel.fake_segment1)
            self.assertIn('hosts', result)
            for host in result['hosts']:
                self.assertFalse(host['on_maintenance'])

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid',
                       return_value=mock.Mock())
    def test_index_invalid_on_maintenance(self, mock_segment):

        req = fakes.HTTPRequest.blank('/v1/segments/%s/hosts?on_maintenance='
                                      'abcd' % uuidsentinel.fake_segment1,
                                      use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req,
                          uuidsentinel.fake_segment1)

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    @mock.patch.object(ha_api.HostAPI, 'get_all')
    def test_index_valid_reserved(self, mock_get_all, mock_segment):
        self.host_list[0]['reserved'] = True
        self.host_list[1]['reserved'] = True
        mock_get_all.return_value = self.host_list
        for parameter in ['1', 't', 'true', 'on', 'y', 'yes']:
            req = fakes.HTTPRequest.blank(
                '/v1/segments/%s/hosts?reserved=''%s' % (
                    uuidsentinel.fake_segment1, parameter
                ), use_admin_context=True)
            result = self.controller.index(req, uuidsentinel.fake_segment1)
            self.assertIn('hosts', result)
            for host in result['hosts']:
                self.assertTrue(host['reserved'])

        self.host_list[0]['reserved'] = False
        self.host_list[1]['reserved'] = False
        mock_get_all.return_value = self.host_list
        for parameter in ['0', 'f', 'false', 'off', 'n', 'no']:
            req = fakes.HTTPRequest.blank(
                '/v1/segments/%s/hosts?reserved=''%s' % (
                    uuidsentinel.fake_segment1, parameter),
                use_admin_context=True)
            result = self.controller.index(req, uuidsentinel.fake_segment1)
            self.assertIn('hosts', result)
            for host in result['hosts']:
                self.assertFalse(host['reserved'])

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid',
                       return_value=mock.Mock())
    def test_index_invalid_reserved(self, mock_segment):

        req = fakes.HTTPRequest.blank('/v1/segments/%s/hosts?reserved='
                                      'abcd' % uuidsentinel.fake_segment1,
                                      use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req,
                          uuidsentinel.fake_segment1)

    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    @mock.patch.object(ha_api.HostAPI, 'get_all')
    def test_index_marker_not_found(self, mock_get_all, mock_segment):
        req = fakes.HTTPRequest.blank('/v1/segments/%s/hosts?marker=123456' % (
            uuidsentinel.fake_segment1), use_admin_context=True)
        mock_segment.return_value = mock.Mock()
        mock_get_all.side_effect = exception.MarkerNotFound(marker="123456")
        self.assertRaises(exc.HTTPBadRequest, self.controller.index,
                          req, uuidsentinel.fake_segment1)

    def test_get_all_marker_negative(self):

        req = fakes.HTTPRequest.blank('/v1/segments/%s/hosts?limit=-1' % (
            uuidsentinel.fake_segment1), use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index,
                          req, uuidsentinel.fake_segment1)

    @ddt.data('sort_key', 'sort_dir')
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid',
                       return_value=mock.Mock())
    def test_index_invalid(self, sort_by, mock_segment):
        req = fakes.HTTPRequest.blank('/v1/segments/%s/hosts?%s=abcd' % (
            uuidsentinel.fake_segment1, sort_by), use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req,
                          uuidsentinel.fake_segment1)

    @ddt.data([exception.MarkerNotFound(marker="123456"),
               "/v1/segments/%s/hosts?marker=123456", exc.HTTPBadRequest],
              [exception.FailoverSegmentNotFound(
                  id=uuidsentinel.fake_segment1), "/v1/segments/%s/hosts",
                  exc.HTTPNotFound])
    @ddt.unpack
    @mock.patch.object(segment_obj.FailoverSegment, 'get_by_uuid')
    @mock.patch.object(ha_api.HostAPI, 'get_all')
    def test_index_not_found(self, masakari_exc, url, exc, mock_get_all,
                             mock_segment):
        mock_segment.return_value = mock.Mock()
        mock_get_all.side_effect = masakari_exc

        req = fakes.HTTPRequest.blank(url % uuidsentinel.fake_segment1,
                                      use_admin_context=True)
        self.assertRaises(exc, self.controller.index, req,
                          uuidsentinel.fake_segment1)

    @mock.patch.object(ha_api.HostAPI, 'create_host')
    def test_create(self, mock_create):
        mock_create.return_value = self.host
        body = {
            "host": {
                "name": "host-1", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"
            }
        }
        result = self.controller.create(self.req,
                                        uuidsentinel.fake_segment1, body=body)
        result = result['host']
        self._assert_host_data(self.host, _make_host_obj(result))

    @mock.patch('masakari.rpc.get_client')
    @mock.patch.object(ha_api.HostAPI, 'create_host')
    def test_create_success_with_201_response_code(
        self, mock_client, mock_create):
        body = {
            "host": {
                "name": "host-1", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"
            }
        }
        fake_req = self.req
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = 'POST'
        fake_req.body = jsonutils.dump_as_bytes(body)
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.CREATED, resp.status_code)

    @mock.patch.object(ha_api.HostAPI, 'create_host')
    def test_create_with_duplicate_host_name(self, mock_create):

        mock_create.side_effect = (exception.
                                   HostExists(name='host-1'))
        body = {
            "host": {
                "name": "host-1", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"
            }
        }
        self.assertRaises(exc.HTTPConflict, self.controller.create,
                          self.req, uuidsentinel.fake_segment1, body=body)

    @ddt.data(
        # no_host
        {"body": {
            "name": "host-1", "type": "fake",
            "reserved": False,
            "on_maintenance": False,
            "control_attributes": "fake-control_attributes"}},

        # no_name
        {"body": {
            "host": {
                "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # name_with_leading_trailing_spaces
        {"body": {
            "host": {
                "name": " host-1 ", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # null_name
        {"body": {
            "host": {
                "name": "", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # name_too_long
        {"body": {
            "host": {
                "name": "host-1" * 255, "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # extra_invalid_arg
        {"body": {
            "host": {
                "name": "host-1", "type": "fake",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes",
                "foo": "bar"}}},

        # type too long
        {"body": {
            "host": {
                "name": "host-1", "type": "x" * 256,
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # type special characters
        {"body": {
            "host": {
                "name": "host-1", "type": "x_y",
                "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}}
    )
    @ddt.unpack
    def test_create_failure(self, body):
        self.assertRaises(self.bad_request, self.controller.create,
                          self.req, uuidsentinel.fake_segment1, body=body)

    @mock.patch.object(ha_api.HostAPI, 'get_host')
    def test_show(self, mock_get_host):

        mock_get_host.return_value = self.host

        result = self.controller.show(self.req, uuidsentinel.fake_segment1,
                                      uuidsentinel.fake_host_1)
        result = result['host']
        self._assert_host_data(self.host, _make_host_obj(result))

    @mock.patch.object(ha_api.HostAPI, 'get_host')
    def test_show_with_non_existing_id(self, mock_get_host):

        mock_get_host.side_effect = exception.HostNotFound(id="2")
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show, self.req,
                          uuidsentinel.fake_segment1, "2")

    @mock.patch.object(ha_api.HostAPI, 'get_host')
    def test_show_non_assigned_failover_segment(self, mock_get_host):

        mock_get_host.side_effect = exception.HostNotFoundUnderFailoverSegment(
            host_uuid=uuidsentinel.fake_host_3,
            segment_uuid=uuidsentinel.fake_segment1)
        self.assertRaises(exc.HTTPNotFound, self.controller.show,
                          self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_3)

    @ddt.data(
        {"body": {
            "host": {
                "name": "host-1", "type": "fake", "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"}}},

        # only name
        {"body": {"host": {"name": "host-1"}}}
    )
    @ddt.unpack
    @mock.patch.object(ha_api.HostAPI, 'update_host')
    def test_update(self, mock_update_host, body):
        mock_update_host.return_value = self.host

        result = self.controller.update(self.req, uuidsentinel.fake_segment1,
                                        uuidsentinel.fake_host_1,
                                        body=body)

        result = result['host']
        self._assert_host_data(self.host, _make_host_obj(result))

    @ddt.data(
        # no updates
        {"test_data": {"host": {}}},

        # no update key
        {"test_data": {"asdf": {}}},

        # wrong updates
        {"test_data": {"host": {"name": "disable", "foo": "bar"}}},

        # null name
        {"test_data": {"host": {"name": ""}}},

        # name too long
        {"test_data": {"host": {"name": "x" * 256}}},

        # type too long
        {"test_data": {"host": {"type": "x" * 256}}},

        # type with special characters
        {"test_data": {"host": {"type": "x_y"}}}
    )
    @ddt.unpack
    def test_update_failure(self, test_data):
        self.assertRaises(self.bad_request, self.controller.update,
                          self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_1, body=test_data)

    @mock.patch.object(ha_api.HostAPI, 'update_host')
    def test_update_with_non_exising_host(self, mock_update_host):

        test_data = {"host": {"name": "host11"}}
        mock_update_host.side_effect = exception.HostNotFound(id="2")
        self.assertRaises(exc.HTTPNotFound, self.controller.update,
                self.req, uuidsentinel.fake_segment1, "2", body=test_data)

    @mock.patch.object(ha_api.HostAPI, 'update_host')
    def test_update_with_duplicated_name(self, mock_update_host):
        test_data = {"host": {"name": "host-1"}}
        mock_update_host.side_effect = exception.HostExists(name="host-1")
        self.assertRaises(exc.HTTPConflict, self.controller.update,
                self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_1, body=test_data)

    @mock.patch.object(ha_api.HostAPI, 'update_host')
    def test_update_non_assigned_failover_segment(self, mock_update_host):
        test_data = {"host": {"name": "host-1"}}
        mock_update_host.side_effect = \
            exception.HostNotFoundUnderFailoverSegment(
                host_uuid=uuidsentinel.fake_host_3,
                segment_uuid=uuidsentinel.fake_segment1)
        self.assertRaises(exc.HTTPNotFound, self.controller.update,
                          self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_3, body=test_data)

    @mock.patch.object(ha_api.HostAPI, 'delete_host')
    def test_delete_host(self, mock_delete):

        self.controller.delete(self.req, uuidsentinel.fake_segment1,
                               uuidsentinel.fake_host_1)
        self.assertTrue(mock_delete.called)

    @mock.patch('masakari.rpc.get_client')
    @mock.patch.object(ha_api.HostAPI, 'delete_host')
    def test_delete_host_with_204_status(self, mock_client, mock_delete):
        url = '/v1/segments/%(segment)s/hosts/%(host)s' % {
            'segment': uuidsentinel.fake_segment1,
            'host': uuidsentinel.fake_host_1
        }
        fake_req = fakes.HTTPRequest.blank(url, use_admin_context=True)
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = 'DELETE'
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.NO_CONTENT, resp.status_code)

    @mock.patch.object(ha_api.HostAPI, 'delete_host')
    def test_delete_host_not_found(self, mock_delete):

        mock_delete.side_effect = exception.HostNotFound(id="2")
        self.assertRaises(exc.HTTPNotFound, self.controller.delete,
                self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_3)

    @mock.patch.object(ha_api.HostAPI, 'delete_host')
    def test_delete_host_not_found_for_failover_segment(self, mock_delete):

        mock_delete.side_effect = exception.HostNotFoundUnderFailoverSegment(
            host_uuid=uuidsentinel.fake_host_3,
            segment_uuid=uuidsentinel.fake_segment1)
        self.assertRaises(exc.HTTPNotFound, self.controller.delete,
                self.req, uuidsentinel.fake_segment1,
                          uuidsentinel.fake_host_3)


class HostTestCasePolicyNotAuthorized(test.NoDBTestCase):
    """Test Case for host non admin."""

    def _set_up(self):
        self.controller = hosts.HostsController()
        self.req = fakes.HTTPRequest.blank(
            '/v1/segments/%s/hosts' % uuidsentinel.fake_segment1)
        self.context = self.req.environ['masakari.context']

    def setUp(self):
        super(HostTestCasePolicyNotAuthorized, self).setUp()
        self._set_up()

    def _check_rule(self, exc, rule_name):
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())

    def test_index_no_admin(self):
        rule_name = "os_masakari_api:os-hosts:index"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.index,
                                self.req, uuidsentinel.fake_segment1)
        self._check_rule(exc, rule_name)

    def test_create_no_admin(self):
        rule_name = "os_masakari_api:os-hosts:create"
        self.policy.set_rules({rule_name: "project:non_fake"})
        body = {
            "host": {
                "name": "host-1", "type": "fake", "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes"
            }
        }
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.create,
                                self.req, uuidsentinel.fake_segment1,
                                body=body)
        self._check_rule(exc, rule_name)

    def test_show_no_admin(self):
        rule_name = "os_masakari_api:os-hosts:detail"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.show,
                                self.req, uuidsentinel.fake_segment1,
                                uuidsentinel.fake_host_1)
        self._check_rule(exc, rule_name)

    def test_update_no_admin(self):
        rule_name = "os_masakari_api:os-hosts:update"
        self.policy.set_rules({rule_name: "project:non_fake"})
        body = {
            "host": {
                "name": "host-1", "type": "fake", "reserved": False,
                "on_maintenance": False,
                "control_attributes": "fake-control_attributes",
            }
        }
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.update,
                                self.req, uuidsentinel.fake_segment1,
                                uuidsentinel.fake_host_1, body=body)
        self._check_rule(exc, rule_name)

    def test_delete_no_admin(self):
        rule_name = "os_masakari_api:os-hosts:delete"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.delete,
                                self.req, uuidsentinel.fake_segment1,
                                uuidsentinel.fake_host_1)
        self._check_rule(exc, rule_name)



#    Copyright 2016 NTT DATA
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_utils import timeutils
from oslo_utils import uuidutils

from masakari.api import utils as api_utils
from masakari import db
from masakari import exception
from masakari.objects import fields
from masakari.objects import notification
from masakari.tests.unit.objects import test_objects
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)
OPTIONAL = ['recovery_workflow_details']


def _fake_db_notification(**kwargs):
    fake_notification = {
        'created_at': NOW,
        'updated_at': None,
        'deleted_at': None,
        'deleted': False,
        'id': 123,
        'notification_uuid': uuidsentinel.fake_notification,
        'generated_time': NOW,
        'type': 'COMPUTE_HOST',
        'payload': '{"fake_key": "fake_value"}',
        'status': 'new',
        'source_host_uuid': uuidsentinel.fake_host,
        }
    fake_notification.update(kwargs)
    return fake_notification


def _fake_object_notification(**kwargs):
    fake_notification = {
        'created_at': NOW,
        'updated_at': None,
        'deleted_at': None,
        'deleted': False,
        'id': 123,
        'notification_uuid': uuidsentinel.fake_notification,
        'generated_time': NOW,
        'type': 'COMPUTE_HOST',
        'payload': {"fake_key": "fake_value"},
        'status': 'new',
        'source_host_uuid': uuidsentinel.fake_host,
        }
    fake_notification.update(kwargs)
    return fake_notification


fake_object_notification = _fake_object_notification()

fake_db_notification = _fake_db_notification()


class TestNotificationObject(test_objects._LocalTest):

    def _test_query(self, db_method, obj_method, *args, **kwargs):
        with mock.patch.object(db, db_method) as mock_db:

            db_exception = kwargs.pop('db_exception', None)
            if db_exception:
                mock_db.side_effect = db_exception
            else:
                mock_db.return_value = fake_db_notification

            obj = getattr(notification.Notification, obj_method
                          )(self.context, *args, **kwargs)
            if db_exception:
                self.assertIsNone(obj)

            self.compare_obj(obj, fake_object_notification,
                             allow_missing=OPTIONAL)

    def test_get_by_id(self):
        self._test_query('notification_get_by_id', 'get_by_id', 123)

    def test_get_by_uuid(self):
        self._test_query('notification_get_by_uuid', 'get_by_uuid',
                         uuidsentinel.fake_segment)

    def _notification_create_attributes(self, skip_uuid=False):

        notification_obj = notification.Notification(context=self.context)
        notification_obj.generated_time = NOW
        notification_obj.type = "COMPUTE_HOST"
        notification_obj.payload = {'fake_key': 'fake_value'}
        notification_obj.status = "new"
        if not skip_uuid:
            notification_obj.notification_uuid = uuidsentinel.fake_notification
        notification_obj.source_host_uuid = uuidsentinel.fake_host

        return notification_obj

    @mock.patch.object(api_utils, 'notify_about_notification_api')
    @mock.patch.object(db, 'notification_create')
    def test_create(self, mock_db_create, mock_notify_about_notification_api):

        mock_db_create.return_value = fake_db_notification
        notification_obj = self._notification_create_attributes()
        notification_obj.create()

        self.compare_obj(notification_obj, fake_object_notification,
                         allow_missing=OPTIONAL)
        mock_db_create.assert_called_once_with(self.context, {
            'source_host_uuid': uuidsentinel.fake_host,
            'notification_uuid': uuidsentinel.fake_notification,
            'generated_time': NOW, 'status': 'new',
            'type': 'COMPUTE_HOST', 'payload': '{"fake_key": "fake_value"}'})
        action = fields.EventNotificationAction.NOTIFICATION_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, notification_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_notification_api')
    @mock.patch.object(db, 'notification_create')
    def test_recreate_fails(self, mock_notification_create,
                            mock_notify_about_notification_api):
        mock_notification_create.return_value = fake_db_notification
        notification_obj = self._notification_create_attributes()
        notification_obj.create()

        self.assertRaises(exception.ObjectActionError, notification_obj.create)

        mock_notification_create.assert_called_once_with(self.context, {
            'source_host_uuid': uuidsentinel.fake_host,
            'notification_uuid': uuidsentinel.fake_notification,
            'generated_time': NOW, 'status': 'new',
            'type': 'COMPUTE_HOST', 'payload': '{"fake_key": "fake_value"}'})
        action = fields.EventNotificationAction.NOTIFICATION_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, notification_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_notification_api')
    @mock.patch.object(db, 'notification_create')
    @mock.patch.object(uuidutils, 'generate_uuid')
    def test_create_without_passing_uuid_in_updates(self, mock_generate_uuid,
                        mock_db_create, mock_notify_about_notification_api):

        mock_db_create.return_value = fake_db_notification
        mock_generate_uuid.return_value = uuidsentinel.fake_notification
        notification_obj = self._notification_create_attributes(skip_uuid=True)

        notification_obj.create()

        self.compare_obj(notification_obj, fake_object_notification,
                         allow_missing=OPTIONAL)
        mock_db_create.assert_called_once_with(self.context, {
            'source_host_uuid': uuidsentinel.fake_host,
            'notification_uuid': uuidsentinel.fake_notification,
            'generated_time': NOW, 'status': 'new',
            'type': 'COMPUTE_HOST', 'payload': '{"fake_key": "fake_value"}'})
        self.assertTrue(mock_generate_uuid.called)
        action = fields.EventNotificationAction.NOTIFICATION_CREATE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, notification_obj, action=action,
                      phase=phase_start)]
        mock_notify_about_notification_api.assert_has_calls(notify_calls)

    @mock.patch.object(db, 'notification_delete')
    def test_destroy(self, mock_notification_delete):
        notification_obj = self._notification_create_attributes()
        notification_obj.id = 123
        notification_obj.destroy()

        (mock_notification_delete.
         assert_called_once_with(self.context, uuidsentinel.fake_notification))
        self.assertRaises(NotImplementedError, lambda: notification_obj.id)

    @mock.patch.object(db, 'notification_delete')
    def test_destroy_without_id(self, mock_destroy):
        notification_obj = self._notification_create_attributes()
        self.assertRaises(exception.ObjectActionError,
                          notification_obj.destroy)
        self.assertFalse(mock_destroy.called)

    @mock.patch.object(db, 'notification_delete')
    def test_destroy_without_notification_uuid(self, mock_destroy):
        notification_obj = self._notification_create_attributes(skip_uuid=True)
        notification_obj.id = 123

        self.assertRaises(exception.ObjectActionError,
                          notification_obj.destroy)
        self.assertFalse(mock_destroy.called)

    @mock.patch.object(db, 'notifications_get_all_by_filters')
    def test_get_notification_by_filters(self, mock_api_get):
        fake_db_notification2 = copy.deepcopy(fake_db_notification)
        fake_db_notification2['type'] = 'PROCESS'
        fake_db_notification2['id'] = 124
        fake_db_notification2[
            'notification_uuid'] = uuidsentinel.fake_db_notification2

        mock_api_get.return_value = [fake_db_notification2,
                                     fake_db_notification]

        filters = {'status': 'new'}
        notification_result = (notification.NotificationList.
                               get_all(self.context, filters=filters))
        self.assertEqual(2, len(notification_result))
        mock_api_get.assert_called_once_with(self.context, filters={
            'status': 'new'
        }, limit=None, marker=None, sort_dirs=None, sort_keys=None)

    @mock.patch.object(db, 'notifications_get_all_by_filters')
    def test_get_limit_and_marker_invalid_marker(self, mock_api_get):
        notification_uuid = uuidsentinel.fake_notification
        mock_api_get.side_effect = (exception.
                                    MarkerNotFound(marker=notification_uuid))

        self.assertRaises(exception.MarkerNotFound,
                          notification.NotificationList.get_all,
                          self.context, limit=5, marker=notification_uuid)

    @mock.patch.object(db, 'notification_update')
    def test_save(self, mock_notification_update):

        mock_notification_update.return_value = fake_db_notification

        notification_obj = self._notification_create_attributes()
        notification_obj.id = 123
        notification_obj.save()

        self.compare_obj(notification_obj, fake_object_notification,
                         allow_missing=OPTIONAL)
        (mock_notification_update.
         assert_called_once_with(self.context, uuidsentinel.fake_notification,
                                 {'source_host_uuid': uuidsentinel.fake_host,
                                  'notificati'
                                  'on_uuid': uuidsentinel.fake_notification,
                                  'status': 'new', 'generated_time': NOW,
                                  'payload': {'fake_key': 'fake_value'},
                                  'type': 'COMPUTE_HOST'}
                                 ))
